"""This module implements the ControlledGate class."""
from __future__ import annotations

from functools import reduce
from typing import Sequence

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.docs import building_docs
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_sequence
from bqskit.utils.typing import is_sequence_of_int


class ControlledGate(ComposedGate, DifferentiableUnitary):
    """
    An arbitrary controlled gate.

    Given any qudit gate, ControlledGate can add control qudits.

    A controlled gate adds arbitrarily controls, and can be generalized
    for qudit or even mixed-qudit representation.

    A controlled gate has a circuit structure as follows:

    ..
        controls ----/----■----
                          |
                         .-.
        targets  ----/---|G|---
                         '-'

    Where `G` is the gate being controlled.

    To calculate the unitary for a controlled gate, given the unitary of
    the gate being controlled, we can use the following equation:

    .. math::
        U_{control} = P_i \\otimes I + P_c \\otimes G

    Where :math:`P_i` is the projection matrix for the states that don't
    activate the gate, :math:`P_c` is the projection matrix for the
    states that do activate the gate, :math:`I` is the identity matrix
    of dimension equal to the gate being controlled, and :math:`G` is
    the unitary matrix of the gate being controlled.

    In the simple case of a normal qubit CNOT, :math:`P_i` and :math:`P_c`
    are defined as follows:

    .. math::

        P_i = |0\\rangle\\langle 0|
        P_c = |1\\rangle\\langle 1|

    This is because the :math:`|0\\rangle` state is the state that doesn't
    activate the gate, and the :math:`|1\\rangle` state is the state that
    does activate the gate.

    We can also decide to invert this, and have the :math:`|0\\rangle`
    state activate the gate, and the :math:`|1\\rangle` state not activate
    the gate. This is equivalent to swapping :math:`P_i` and :math:`P_c`,
    and usually drawn diagrammatically as follows:

    ..
        controls ----/----□----
                          |
                         .-.
        targets  ----/---|G|---
                         '-'


    When we add more controls the projection matrices become more complex,
    but the basic idea stays the same: we have a projection matrix for
    the states that activate the gate, and a projection matrix for the
    states that don't activate the gate. As in the case of a toffoli gate,
    the projection matrices are defined as follows:

    .. math::

        P_i = |00\\rangle\\langle 00| + |01\\rangle\\langle 01|
            + |10\\rangle\\langle 10|

        P_c = |11\\rangle\\langle 11|

    This is because the :math:`|00\\rangle`, :math:`|01\\rangle`, and
    :math:`|10\\rangle` states are the states that don't activate the
    gate, and the :math:`|11\\rangle` state is the state that does
    activate the gate.

    With qudits, we have more states and as such, more complex
    projection matrices; however, the basic idea is the same.
    For example, a qutrit controlled-not gate that is activated by
    the :math:`|2\\rangle` state and not activated by the :math:`|0\\rangle`
    and :math:`|1\\rangle` states is defined as follows:

    .. math::

        P_i = |0\\rangle\\langle 0| + |1\\rangle\\langle 1|
        P_c = |2\\rangle\\langle 2|

    One interesting concept with qudits is that we can have multiple
    active control levels. For example, a qutrit controlled-not gate that
    is activated by the :math:`|1\\rangle` and :math:`|2\\rangle` states
    and not activated by the :math:`|0\\rangle` state is defined similarly
    as follows:

    .. math::

        P_i = |0\\rangle\\langle 0|
        P_c = |1\\rangle\\langle 1| + |2\\rangle\\langle 2|

    Note that we can always define :math:`P_i` simply from :math:`P_c`:

    .. math::

        P_i = I_p - P_c

    Where :math:`I_p` is the identity matrix of dimension equal to the
    dimension of the control qudits. This leaves us with out final
    equation:

    .. math::

        U_{control} = (I_p - P_c) \\otimes I + P_c \\otimes G

    If, G is a unitary-valued function of real parameters, then the
    gradient of the controlled gate simply discards the constant half
    of the equation:

    .. math::

        \\frac{\\partial U_{control}}{\\partial \\theta} =
            P_c \\otimes \\frac{\\partial G}{\\partial \\theta}
    """

    def __init__(
        self,
        gate: Gate,
        num_controls: int = 1,
        control_radixes: Sequence[int] | int = 2,
        control_levels: Sequence[Sequence[int] | int] | int | None = None,
    ):
        """
        Construct a ControlledGate.

        Args:
            gate (Gate): The gate to control.

            num_controls (int): The number of controls.

            control_radixes (Sequence[int] | int): The number of levels
                for each control qudit. If one number is provided, all
                control qudits will have the same number of levels.
                Defaults to qubits (2).

            control_levels  (Sequence[Sequence[int] | int] | int | None):
                Sequence of control levels for each control qudit. These
                levels need to be activated on the corresponding control
                qudits for the gate to be activated. If more than one
                level is selected, the subspace spanned by the levels
                acts as a control subspace. If all levels are selected
                for a given qudit, the operation is equivalent to the
                original gate without controls. If None, the highest
                level acts as the control level for each control qudit.
                Can be given as a single integer, which will be broadcast
                as the control level for all control qudits; or as a
                sequence, where each element describes the control levels
                for the corresponding control qudit. These can be given
                as a single integer or a sequence of integers. Defaults
                to None.

        Raises:
            ValueError: If `num_controls` is less than 1

            ValueError: If any control radix is less than 2

            ValueError: If any control level is less than 0

            ValueError: If the length of control_radixes is not equal to
                num_controls

            ValueError: If the length of control_levels is not equal to
                num_controls

            ValueError: If any control level is repeated in for the same
                qudit

            ValueError: If there exists an `i` and `j` where
                `control_levels[i][j] >= control_radixes[i]`

        Examples:
            If we didn't have the CNOTGate we can do it from this gate:

            >>> from bqskit.ir.gates import XGate, ControlledGate
            >>> cnot_gate = ControlledGate(XGate())
            >>> cnot_gate.get_unitary()
            array([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                   [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                   [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
                   [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])

            We can invert the controls of a CNOTGate, by activating on the
            |0> state instead of the |1> state:

            >>> from bqskit.ir.gates import XGate, ControlledGate
            >>> inverted_cnot_gate = ControlledGate(XGate(), control_levels=0)

            Also, if we didn't have the ToffoliGate for qubits, we could do:

            >>> from bqskit.ir.gates import XGate, ControlledGate, ToffoliGate
            >>> toffoli_gate = ControlledGate(XGate(), 2) # 2 controls
            >>> ToffoliGate().get_unitary() == toffoli_gate.get_unitary()
            True

            We can define a qutrit CNOT that is activated by the |2> state:

            >>> from bqskit.ir.gates import ShiftGate, ControlledGate
            >>> qutrit_x = ShiftGate(3)
            >>> qutrit_cnot = ControlledGate(qutrit_x, control_radixes=3)

            This composed gate can also be used to define mixed-radix
            controlled systems. For example, we can define a mixed-radix
            CNOT where the control is a qubit and the X Gate is qutrit:

            >>> from bqskit.ir.gates import ShiftGate, ControlledGate
            >>> qutrit_x = ShiftGate(3)
            >>> qutrit_cnot = ControlledGate(qutrit_x)

            We can also define multiple controls with mixed qudits
            that require multiple levels to be activated. In this
            example, The first control is a qutrit with [0,1] control
            levels, the second qudit is a ququart with a [0] control level,
            and RY Gate for qubit operation:

            >>> from bqskit.ir.gates import RYGate, ControlledGate
            >>> cgate = ControlledGate(
            ...     RYGate(),
            ...     num_controls=2,
            ...     control_radixes=[3,4],
            ...     control_levels=[[0,1], 0]
            ... )
        """
        if not isinstance(gate, Gate):
            raise TypeError(f'Expected gate object, got {type(gate)}.')

        self.gate = gate

        params = self._check_and_type_control_parameters(
            num_controls,
            control_radixes,
            control_levels,
        )
        num_controls, control_radixes, control_levels = params
        self.num_controls, self.control_radixes, self.control_levels = params

        self._radixes = tuple(tuple(control_radixes) + self.gate.radixes)
        self._num_qudits = gate._num_qudits + self.num_controls
        # TODO: Incorporate control radixes/levels into name with function def.
        self._name = 'Controlled(%s)' % self.gate.name
        self._num_params = self.gate._num_params

        iden_gate = np.identity(self.gate.dim, dtype=np.complex128)
        """Identity is applied when controls are not properly activated."""

        self.ctrl = self.build_control_proj(control_radixes, control_levels)
        """Control projection matrix determines if the gate should activate."""

        ctrl_dim = int(np.prod(self.control_radixes))
        """Dimension of the control qudits."""

        iden_proj = np.eye(ctrl_dim, dtype=np.complex128) - self.ctrl
        """Identity projection matrix determines if it shouldn't activate."""

        self.ihalf = np.kron(iden_proj, iden_gate)
        """Identity half of the final unitary equation."""

        # If input is a constant gate, we can cache the unitary.
        if self.num_params == 0 and not building_docs():
            U = self.gate.get_unitary()
            ctrl_U = np.kron(self.ctrl, U) + self.ihalf
            self._utry = UnitaryMatrix(ctrl_U, self.radixes)

    @property
    def qasm_name(self) -> str:
        """
        Override default `Gate.qasm_name` method.

        If the core gate is a standard gate, this function will output
        qasm in the form 'c+<gate_qasm>'. Otherwise an error will be raised.

        Raises:
            ValueError: If the core gate is non-standard in OpenQASM 2.0.
        """
        _core_gate = self.gate.qasm_name
        if self.num_controls <= 2:
            _controls = 'c' * self.num_controls
        else:
            _controls = f'c{self.num_controls}'
        qasm_name = _controls + _core_gate
        supported_gates = ('cu1', 'cu2', 'cu3', 'cswap', 'c3x', 'c4x')
        if qasm_name not in supported_gates:
            raise ValueError(
                f'Controlled gate {_core_gate} with {self.num_controls} '
                'controls is not a standard OpenQASM 2.0 identifier. '
                'To encode this gate, try decomposing it into gates with'
                'standard identifiers.',
            )
        return qasm_name

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        if hasattr(self, '_utry'):
            return self._utry

        U = self.gate.get_unitary(params)
        ctrl_U = np.kron(self.ctrl, U) + self.ihalf
        return UnitaryMatrix(ctrl_U, self.radixes)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, '_utry'):
            return np.array([])

        grads = self.gate.get_grad(params)  # type: ignore
        return np.kron(self.ctrl, grads)

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, '_utry'):
            return self._utry, np.array([])

        U, grads = self.gate.get_unitary_and_grad(params)  # type: ignore
        ctrl_U = np.kron(self.ctrl, U) + self.ihalf
        ctl_grads = np.kron(self.ctrl, grads)
        return UnitaryMatrix(ctrl_U, self.radixes), ctl_grads

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ControlledGate)
            and self.gate == other.gate
            and self.num_controls == other.num_controls
            and self.control_radixes == other.control_radixes
            and self.control_levels == other.control_levels
        )

    def __hash__(self) -> int:
        return hash((self.gate, self.radixes))

    @staticmethod
    def build_control_proj(
        control_radixes: Sequence[int],
        control_levels: Sequence[Sequence[int]],
    ) -> npt.NDArray[np.complex128]:
        """
        Construct the control projection matrix from the control levels.

        See :class:`ControlledGate` for more info.
        """
        params = ControlledGate._check_and_type_control_parameters(
            len(control_radixes),
            control_radixes,
            control_levels,
        )
        _, control_radixes, control_levels = params

        elementary_projection_list = [
            np.zeros((r, r), dtype=np.complex128)
            for r in control_radixes
        ]

        for i, projection in enumerate(elementary_projection_list):
            for control_level in control_levels[i]:
                projection[control_level, control_level] = 1.0

        return reduce(np.kron, elementary_projection_list)  # type: ignore

    @staticmethod
    def _check_and_type_control_parameters(
        num_controls: int = 1,
        control_radixes: Sequence[int] | int = 2,
        control_levels: Sequence[Sequence[int] | int] | int | None = None,
    ) -> tuple[int, list[int], list[list[int]]]:
        """
        Checks the control paramters for type and value errors.

        Returns specific types for each parameter; see
        :class:`ControlledGate` for more info on errors and parameters.
        """
        if not is_integer(num_controls):
            raise TypeError(
                f'Expected integer for num_controls, got {type(num_controls)}.',
            )

        if num_controls < 1:
            raise ValueError(
                'Expected num_controls to be greater than or equal to 1'
                f', got {num_controls}.',
            )

        if is_integer(control_radixes):
            control_radixes = [control_radixes] * num_controls

        if not is_sequence_of_int(control_radixes):
            raise TypeError(
                'Expected integer or sequence of integers for control_radixes.',
            )

        if any(r < 2 for r in control_radixes):
            raise ValueError('Expected every radix to be greater than 1.')

        if len(control_radixes) != num_controls:
            raise ValueError(
                'Expected length of control_radixes to be equal to '
                f'num_controls: {len(control_radixes)=} != {num_controls=}.',
            )

        if control_levels is None:
            control_levels = [
                [control_radixes[i] - 1]
                for i in range(num_controls)
            ]

        if is_integer(control_levels):
            control_levels = [[control_levels]] * num_controls

        if not is_sequence(control_levels):
            raise TypeError(
                'Expected sequence of sequence of integers for control_levels,'
                f'got {type(control_levels)}.',
            )

        control_levels = [
            [level] if is_integer(level) else level
            for level in control_levels
        ]

        if any(not is_sequence_of_int(levels) for levels in control_levels):
            bad = [not is_sequence_of_int(levels) for levels in control_levels]
            bad_index = bad.index(True)
            raise TypeError(
                'Expected sequence of sequence of integers for control_levels,'
                f'got {control_levels[bad_index]} where a sequence of integers'
                ' was expected.',
            )

        if len(control_levels) != num_controls:
            raise ValueError(
                'Expected length of control_levels to be equal to '
                f'num_controls: {len(control_levels)=} != {num_controls=}.',
            )

        if any(l < 0 for levels in control_levels for l in levels):
            raise ValueError(
                'Expected control levels to be greater than or equal to 0.',
            )

        if any(
            l >= r
            for r, levels in zip(control_radixes, control_levels)
            for l in levels
        ):
            raise ValueError(
                'Expected control levels to be less than the number of levels.',
            )

        if any(len(l) != len(set(l)) for l in control_levels):
            raise ValueError(
                'Expected control levels to be unique for each qudit.',
            )

        control_levels = [list(levels) for levels in control_levels]
        return num_controls, list(control_radixes), control_levels



"""This module implements the SubSwapGate."""
from __future__ import annotations

import numpy as np

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class SubSwapGate(ConstantGate, QuditGate):
    """
    The two-qudit subspace SWAP gate.

    The subspace SWAP gate swaps between "qudit-levels" on a two-qudit gate. For
    example, a |01> to |20> swap would be the identity with the |01> row and
    |20> rows swapped.
    """

    _num_qudits = 2
    _num_params = 0

    def __init__(self, radix: int, qudit_levels: str) -> None:
        """
        Construct a SubSwapGate.

        Args:
            radix (int): The number of qudit levels (>=2).

            qudit_levels (str): The qudit levels that should be swapped,
                separated by a comma. Example: "0,1;2,0" to swap |01>
                to |20>

        Raises:
            ValueError: if radix < 2

            ValueError: If any of the qudit levels integer represenation
                greater than or equal to radix.

            ValueError: If the qudit_levels string is not formatted correctly.
        """

        if not is_integer(radix):
            raise TypeError(f'Expected integer for radix, got {type(radix)}.')

        if radix < 2:
            raise ValueError(f'Radix must be greater than 1, got {radix}.')

        if not isinstance(qudit_levels, str):
            raise TypeError(
                'Expected qudit_levels object to be string'
                f', got {type(qudit_levels)}.',
            )

        self._radix = radix
        level1, level2 = self.decode_qudit_level_string(qudit_levels)
        self._utry = self.calculate_level_swap_unitary(radix, level1, level2)

    @staticmethod
    def calculate_level_swap_unitary(
        radix: int,
        level1: tuple[int, int],
        level2: tuple[int, int],
    ) -> UnitaryMatrix:
        """
        Calculate the unitary for a qudit level swap.

        Args:
            radix (int): The number of qudit levels (>=2).

            level1 (tuple[int, int]): The first qudit level to swap.

            level2 (tuple[int, int]): The second qudit level to swap.

        Returns:
            UnitaryMatrix: The unitary matrix for the qudit level swap.

        Raises:
            ValueError: If any of the qudit levels integer represenation
                greater than or equal to radix.

        Example:
            >>> from bqskit.ir.gates.constant.subswap import SubSwapGate
            >>> u = SubSwapGate.calculate_level_swap_unitary(2, (1, 1), (1, 0))
            >>> u
            array([[1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                   [0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
                   [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
                   [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])
            >>> from bqskit.ir.gates.constant.cx import CNOTGate
            >>> u == CNOTGate().get_unitary()
            True
        """
        if any(l >= radix for l in level1 + level2):
            raise ValueError(
                'Qudit levels must be less than radix'
                f', got |{level1}> swaps |{level2}> when {radix=}.',
            )

        utry = np.identity(radix ** 2, dtype=np.complex128)
        i = level1[0] * radix + level1[1]
        j = level2[0] * radix + level2[1]
        utry[i, i] = 0
        utry[j, j] = 0
        utry[i, j] = 1
        utry[j, i] = 1
        return UnitaryMatrix(utry, [radix, radix])

    @staticmethod
    def decode_qudit_level_string(
        string: str,
    ) -> tuple[tuple[int, int], tuple[int, int]]:
        """
        Decode the qudit level string into two pairs of integers.

        Args:
            string (str): The qudit level string to decode. See the
                :class:`SubSwapGate` documentation for more info on the
                format.

        Returns:
            tuple[tuple[int, int], tuple[int, int]]: The two two-qudit
                levels to swap, i.e., "0,1;2,0" to swap |01> to |20>.

        Raises:
            ValueError: See :class:`SubSwapGate` documentation for more info.
        """
        split_values = string.split(';', 1)

        if len(split_values) != 2:
            raise ValueError(
                'Qudit levels string must contain exactly one semicolon,'
                f' got {string}.',
            )

        first = split_values[0].split(',')
        second = split_values[1].split(',')

        if len(first) != 2 or len(second) != 2:
            raise ValueError(
                'Qudit levels string must contain exactly one comma in each'
                f' section, got {string}.',
            )

        return (int(first[0]), int(first[1])), (int(second[0]), int(second[1]))



"""This module implements the MPRZGate."""
from __future__ import annotations

import typing
from typing import Sequence

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.parameterized.mpry import get_indices
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class MPRZGate(
    QubitGate,
    DifferentiableUnitary,
    CachedClass,
    LocallyOptimizableUnitary,
):
    """
    A gate representing a multiplexed Z rotation. A multiplexed Z rotation
    uses n - 1 qubits as select qubits and applies a Z rotation to the target.
    If the target qubit is the last qubit, then the unitary is block diagonal.
    Each block is a 2x2 RZ matrix with parameter theta.

    Since there are n - 1 select qubits, there are 2^(n-1) parameters (thetas).

    We allow the target qubit to be specified to any qubit, and the other qubits
    maintain their order. Qubit 0 is the most significant qubit.


    Why is 0 the MSB? Typically, in the QSD diagram, we see the block drawn
    with qubit 0 at the top and qubit n-1 at the bottom. Then, the decomposition
    slowly moves from the bottom to the top.

    See this paper: https://arxiv.org/pdf/quant-ph/0406176
    """

    _qasm_name = 'mprz'

    def __init__(
        self,
        num_qudits: int,
        target_qubit: int = -1,
    ) -> None:
        """
        Create a new MPRZGate with `num_qudits` qubits and `target_qubit` as the
        target qubit. We then have 2^(n-1) parameters for this gate.

        For Example:
        `num_qudits` = 3, `target_qubit` = 1

        Then, the select qubits are 0 and 2 with 0 as the MSB.

        If the input vector is |0x0> then the selection is 00, and
        RZ(theta_0) is applied to the target qubit.

        If the input vector is |1x0> then the selection is 01, and
        RZ(theta_1) is applied to the target qubit.
        """
        self._num_qudits = num_qudits
        # 1 param for each configuration of the selec qubits
        self._num_params = 2 ** (num_qudits - 1)
        # By default, the controlled qubit is the last qubit
        if target_qubit == -1:
            target_qubit = num_qudits - 1
        self.target_qubit = target_qubit
        super().__init__()

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)
        matrix = np.zeros(
            (
                2 ** self.num_qudits,
                2 ** self.num_qudits,
            ), dtype=np.complex128,
        )
        for i, param in enumerate(typing.cast(typing.Sequence[float], params)):
            pos = np.exp(1j * param / 2)
            neg = np.exp(-1j * param / 2)

            # Get correct indices based on target qubit
            # See :class:`mcry` for more info
            x1, x2 = get_indices(i, self.target_qubit, self.num_qudits)

            matrix[x1, x1] = neg
            matrix[x2, x2] = pos

        return UnitaryMatrix(matrix)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        grad = np.zeros(
            (
                len(params), 2 ** self.num_qudits,
                2 ** self.num_qudits,
            ), dtype=np.complex128,
        )

        # For each parameter, calculate the derivative
        # with respect to that parameter
        for i, param in enumerate(typing.cast(Sequence[float], params)):
            dpos = 1j * np.exp(1j * param / 2) / 2
            dneg = -1j * np.exp(-1j * param / 2) / 2

            # Again, get indices based on target qubit.
            x1, x2 = get_indices(i, self.target_qubit, self.num_qudits)

            grad[i, x1, x1] = dpos
            grad[i, x2, x2] = dneg

        return grad

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        self.check_env_matrix(env_matrix)
        thetas: list[float] = [0] * self.num_params

        for i in range(self.num_params):
            x1, x2 = get_indices(i, self.target_qubit, self.num_qudits)
            # Optimize each RZ independently from indices
            # Taken from QFACTOR repo
            a = np.angle(env_matrix[x1, x1])
            b = np.angle(env_matrix[x2, x2])
            # print(thetas)
            thetas[i] = a - b

        return thetas

    @property
    def name(self) -> str:
        """The name of this gate, with the number of qudits appended."""
        base_name = getattr(self, '_name', self.__class__.__name__)
        return f'{base_name}_{self.num_qudits}'



"""This module implements the CircuitRegion class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Iterator
from typing import Mapping
from typing import Union

from typing_extensions import TypeGuard

from bqskit.ir.interval import CycleInterval
from bqskit.ir.interval import IntervalLike
from bqskit.ir.location import CircuitLocation
from bqskit.ir.point import CircuitPoint
from bqskit.ir.point import CircuitPointLike
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_mapping
_logger = logging.getLogger(__name__)


class CircuitRegion(Mapping[int, CycleInterval]):
    """
    The CircuitRegion class.

    A CircuitRegion is an contiguous, convex area in a circuit. It is
    represented as a map from qudit indices to cycle intervals.
    """

    def __init__(self, intervals: Mapping[int, IntervalLike]) -> None:
        """
        CircuitRegion Initializer.

        Args:
            intervals (Mapping[int, IntervalLike]): A map from qudit
                indices to cycle intervals. The cycle intervals can
                be given as either a CycleInterval object or a tuple
                of two ints that represent the lower and upper bound
                of the inclusive interval.

        Notes:
            All cycle intervals are inclusive.
        """
        if not is_mapping(intervals):
            raise TypeError(
                f'Expected mapping from int to IntervalLike, got {intervals}.',
            )

        for qudit, interval in intervals.items():
            if not is_integer(qudit):
                raise TypeError(f'Expected integer keys, got {qudit}.')

            if not CycleInterval.is_interval(interval):
                raise TypeError(
                    f'Expected valid CycleInterval, got {interval}.',
                )

        self._intervals = {
            qudit: CycleInterval(interval)
            for qudit, interval in intervals.items()
        }

    def __getitem__(self, key: int) -> CycleInterval:
        """
        Return the interval given by the qudit `key`.

        Raises:
            IndexError: If `key` is not in this region.
        """
        return self._intervals[key]

    def __iter__(self) -> Iterator[int]:
        """Return an iterator for all qudits contained in the region."""
        return iter(self._intervals)

    def __len__(self) -> int:
        """Return the number of qudits in the region."""
        return len(self._intervals)

    @property
    def min_cycle(self) -> int:
        """
        Return the smallest cycle index for any qudit.

        Raises:
            ValueError: If `self` is empty.
        """
        if self.empty:
            raise ValueError('Empty region cannot have minimum cycle.')

        return min(interval.lower for interval in self.values())

    @property
    def max_cycle(self) -> int:
        """
        Return the largest cycle index for any qudit.

        Raises:
            ValueError: If `self` is empty.
        """
        if self.empty:
            raise ValueError('Empty region cannot have maximum cycle.')

        return max(interval.upper for interval in self.values())

    @property
    def max_min_cycle(self) -> int:
        """
        Return the largest cycle index for any lower bound.

        Raises:
            ValueError: If `self` is empty.
        """
        if self.empty:
            raise ValueError('Empty region cannot have minimum cycle.')

        return max(interval.lower for interval in self.values())

    @property
    def min_max_cycle(self) -> int:
        """
        Return the smallest cycle index for any upper bound.

        Raises:
            ValueError: If `self` is empty.
        """
        if self.empty:
            raise ValueError('Empty region cannot have maximum cycle.')

        return min(interval.upper for interval in self.values())

    @property
    def min_qudit(self) -> int:
        """
        Return the smallest qudit index.

        Raises:
            ValueError: If `self` is empty.
        """
        if self.empty:
            raise ValueError('Empty region cannot have minimum qudit.')

        return min(self.keys())

    @property
    def max_qudit(self) -> int:
        """
        Return the largest qudit index.

        Raises:
            ValueError: If `self` is empty.
        """
        if self.empty:
            raise ValueError('Empty region cannot have maximum qudit.')

        return max(self.keys())

    @property
    def location(self) -> CircuitLocation:
        """Return the qudits this region includes as a CircuitLocation."""
        return CircuitLocation(sorted(self.keys()))

    @property
    def points(self) -> list[CircuitPoint]:
        """Return the points described by this region."""
        return [
            CircuitPoint(cycle_index, qudit_index)
            for qudit_index, intervals in self.items()
            for cycle_index in intervals.indices
        ]

    @property
    def volume(self) -> int:
        """Return the volume of this region measured in qudit-cycles."""
        return sum(len(interval) for interval in self.values())

    @property
    def width(self) -> int:
        """Return the number of cycles this region participates in."""
        if self.empty:
            return 0

        return self.max_cycle - self.min_cycle + 1

    @property
    def empty(self) -> bool:
        """Return true if this region is empty."""
        return len(self) == 0

    @property
    def num_qudits(self) -> int:
        """Return the number of qudits in this region."""
        return len(self)

    def shift_left(self, amount_to_shift: int) -> CircuitRegion:
        """
        Shift the region to the left by `amount_to_shift`.

        Args:
            amount_to_shift (int): Subtract `amount_to_shift` from all
                cycle indices.

        Raises:
            ValueError: If the region would be shifted below zero as a
                result of performing this operation.
        """
        if not is_integer(amount_to_shift):
            raise TypeError(
                f'Expected integer for shift amount, got {amount_to_shift}.',
            )

        if self.empty:
            return CircuitRegion(self._intervals)

        if self.min_cycle - amount_to_shift < 0:
            raise ValueError(
                f'Cannot shift region to the left by {amount_to_shift}.',
            )

        return CircuitRegion({
            qudit_index:
            CycleInterval(
                interval[0] - amount_to_shift,
                interval[1] - amount_to_shift,
            )
            for qudit_index, interval in self._intervals.items()
        })

    def shift_right(self, amount_to_shift: int) -> CircuitRegion:
        """
        Shift the region to the right by `amount_to_shift`.

        Args:
            amount_to_shift (int): Add `amount_to_shift` to all
                cycle indices.
        """
        if not is_integer(amount_to_shift):
            raise TypeError(
                f'Expected integer for shift amount, got {amount_to_shift}.',
            )

        if self.empty:
            return CircuitRegion(self._intervals)

        if amount_to_shift < 0:
            return self.shift_left(-amount_to_shift)

        return CircuitRegion({
            qudit_index:
            CycleInterval(
                interval[0] + amount_to_shift,
                interval[1] + amount_to_shift,
            )
            for qudit_index, interval in self._intervals.items()
        })

    def overlaps(self, other: CircuitPointLike | CircuitRegionLike) -> bool:
        """Return true if `other` overlaps this region."""

        if CircuitPoint.is_point(other):
            other = CircuitPoint(*other)

            if other.qudit not in self:
                return False

            intervals = self[other.qudit]
            return intervals.lower <= other.cycle <= intervals.upper

        if CircuitRegion.is_region(other):
            other = CircuitRegion(other)

            if other.empty or self.empty:
                return False

            if self.min_cycle > other.max_cycle:
                return False

            if self.max_cycle < other.min_cycle:
                return False

            qudit_intersection = self.location.intersection(other.location)

            if len(qudit_intersection) == 0:
                return False

            for qudit in qudit_intersection:
                if (
                    self[qudit].lower <= other[qudit].upper
                    and self[qudit].upper >= other[qudit].lower
                ):
                    return True

            return False

        raise TypeError(
            'Expected either CircuitPoint or CircuitRegion, got %s.'
            % type(other),
        )

    def copy(self) -> CircuitRegion:
        """Return a deep copy of this region."""
        return CircuitRegion(self._intervals)

    def __contains__(self, other: object) -> bool:
        if is_integer(other):
            return other in self._intervals.keys()

        if CircuitPoint.is_point(other):
            return other[1] in self.keys() and other[0] in self[other[1]]

        if CircuitRegion.is_region(other):
            other = CircuitRegion(other)

            if other.empty:
                return True

            if self.empty:
                return False

            return (
                all(qudit in self.keys() for qudit in other.keys())
                and all(
                    self[qudit].lower <= other[qudit][0] <= self[qudit].upper
                    for qudit in other.keys()
                )
                and all(
                    self[qudit].lower <= other[qudit][1] <= self[qudit].upper
                    for qudit in other.keys()
                )
            )

        return NotImplemented

    def transpose(self) -> dict[int, list[int]]:
        """Flip region to map cycle indices to qudit indices."""
        if self.empty:
            return {}

        qudit_cycles: dict[int, list[int]] = {
            i: []
            for i in range(self.min_cycle, self.max_cycle + 1)
        }

        for qudit_index, intervals in sorted(self.items()):
            for cycle_index in range(intervals.lower, intervals.upper + 1):
                qudit_cycles[cycle_index].append(qudit_index)

        return {k: v for k, v in qudit_cycles.items() if len(v) > 0}

    def intersection(self, other: CircuitRegionLike) -> CircuitRegion:
        if not CircuitRegion.is_region(other):
            raise TypeError(f'Expected CircuitRegion, got {type(other)}.')

        other = CircuitRegion(other)
        location = self.location.intersection(other.location)

        region: dict[int, CycleInterval] = {}
        for qudit in location:
            if self[qudit].overlaps(other[qudit]):
                region[qudit] = self[qudit].intersection(other[qudit])

        return CircuitRegion(region)

    def union(self, other: CircuitRegionLike) -> CircuitRegion:
        if not CircuitRegion.is_region(other):
            raise TypeError(f'Expected CircuitRegion, got {type(other)}.')

        other = CircuitRegion(other)
        location = self.location.union(other.location)

        region: dict[int, CycleInterval] = {}
        for qudit in location:
            if qudit in self and qudit in other:
                region[qudit] = self[qudit].union(other[qudit])

            elif qudit in self:
                region[qudit] = self[qudit]

            elif qudit in other:
                region[qudit] = other[qudit]

        return CircuitRegion(region)

    def depends_on(self, other: CircuitRegionLike) -> bool:
        """Return true if self depends on other."""
        if not isinstance(other, CircuitRegion):
            other = CircuitRegion(other)

        intersection = self.location.intersection(other.location)
        if len(intersection) != 0:
            return all(other[qudit] < self[qudit] for qudit in intersection)

        return False

    def dependency(self, other: CircuitRegionLike) -> int:
        """
        Return 1 if self depends on other.

        Return -1 if other depends on self. Return 0 if no dependency.
        """

        if not isinstance(other, CircuitRegion):
            other = CircuitRegion(other)

        intersection = self.location.intersection(other.location)

        if not intersection:
            return 0

        for qudit in intersection:
            if other[qudit] < self[qudit]:
                return 1

        return -1

    def __eq__(self, other: object) -> bool:
        if CircuitRegion.is_region(other):
            other = CircuitRegion(other)
            return sorted(self.items()) == sorted(other.items())

        return NotImplemented

    def __hash__(self) -> int:
        return tuple(sorted(self.items())).__hash__()

    def __str__(self) -> str:
        return str(self._intervals)

    def __repr__(self) -> str:
        return repr(self._intervals)

    def __lt__(self, other: object) -> bool:
        if CircuitPoint.is_point(other):
            other = CircuitPoint(*other)
            if other[0] < self.min_cycle:
                return True

            if other[1] in self.keys():
                return other[0] < self[other[1]].lower

        elif CircuitRegion.is_region(other):
            other = CircuitRegion(other)

            if len(self.location.intersection(other.location)) != 0:
                lt = None
                for qudit in self.location.intersection(other.location):
                    if lt is None:
                        lt = self[qudit] < other[qudit]
                    elif lt != (self[qudit] < other[qudit]):
                        raise ValueError('Both regions depend on each other.')

                assert lt is not None
                return lt

            lower_intervals = tuple(sorted({x.lower for x in self.values()}))
            other_lower_intervals = tuple(
                sorted({x.lower for x in other.values()}),
            )
            upper_intervals = tuple(
                reversed(sorted({x.upper for x in self.values()})),
            )
            other_upper_intervals = tuple(
                reversed(sorted({x.upper for x in other.values()})),
            )
            return (lower_intervals, upper_intervals) < (
                other_lower_intervals, other_upper_intervals,
            )

        return NotImplemented

    @staticmethod
    def is_region(region: Any) -> TypeGuard[CircuitRegionLike]:
        """Return true if region is a CircuitRegionLike."""
        if isinstance(region, CircuitRegion):
            return True

        if not is_mapping(region):
            _logger.log(0, 'Region is not a mapping.')
            return False

        if not all(is_integer(key) for key in region.keys()):
            _logger.log(0, 'Region does not have integer keys.')
            return False

        if not all(CycleInterval.is_interval(val) for val in region.values()):
            _logger.log(0, 'Region does not have IntervalLike values.')
            return False

        return True


CircuitRegionLike = Union[Mapping[int, IntervalLike], CircuitRegion]



"""This module implements the ECRGate."""
from __future__ import annotations

import math

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class ECRGate(ConstantGate, QubitGate):
    """
    The echoed cross-resonance gate (ECR).

    The ECR gate is given by the following unitary:

    .. math::

        \\frac{1}{\\sqrt{2}}
        \\begin{pmatrix}
        0  & 1 & 0  & i \\\\
        1  & 0 & -i & 0 \\\\
        0  & i & 0  & 1 \\\\
        -i & 0 & 1  & 0
        \\end{pmatrix}
    """
    _name = 'ecr'
    _num_qudits = 2
    _qasm_name = 'ecr'
    _utry = UnitaryMatrix([
        [0, 0, 1 * 1 / math.sqrt(2), 1j * 1 / math.sqrt(2)],
        [0, 0, 1j * 1 / math.sqrt(2), 1 * 1 / math.sqrt(2)],
        [1 * 1 / math.sqrt(2), -1j * 1 / math.sqrt(2), 0, 0],
        [-1j * 1 / math.sqrt(2), 1 * 1 / math.sqrt(2), 0, 0],
    ])

    def __init__(self) -> None:
        pass

    def get_qasm_gate_def(self) -> str:
        qasm_rzx = (
            'gate rzx(param0) q0,q1 {'
            'h q1;'
            'cx q0,q1;'
            'rz(param0) q1;'
            'cx q0,q1;'
            'h q1; }'
        )
        qasm_ecr = 'gate ecr q0,q1 { rzx(pi/4) q0,q1; x q0; rzx(-pi/4) q0,q1; }'
        return qasm_rzx + '\n' + qasm_ecr



"""This module implements the StairLayerGenerator class."""
from __future__ import annotations

import logging

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


_logger = logging.getLogger(__name__)


class StairLayerGenerator(LayerGenerator):
    """Layer Generator for search that builds circuits from a single gate."""

    def __init__(self, gate: Gate) -> None:
        """
        Construct a StairLayerGenerator.

        Args:
            gate (Gate): The gate to build from.
        """
        if not isinstance(gate, Gate):
            raise TypeError(f'Expected Gate for gate, got {type(gate)}')

        self.gate = gate

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a size or radix mismatch with
                `self.seed`.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        init_circuit = Circuit(target.num_qudits, target.radixes)

        # Place RXX Gates on consective pairs of qudits
        nq = self.gate.num_qudits
        for i in range(init_circuit.num_qudits - (nq - 1)):
            init_circuit.append_gate(self.gate, [i + j for j in range(nq)])

        return init_circuit

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        nq = self.gate.num_qudits
        successor = circuit.copy()
        for i in range(circuit.num_qudits - (nq - 1)):
            successor.append_gate(self.gate, [i + j for j in range(nq)])

        return [successor]



"""This module implements the PDGate."""
from __future__ import annotations

import numpy as np

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class PDGate(ConstantGate, QuditGate):
    """
    The one-qudit P[i] gate.

    The clock gate is given by the following formula:

    .. math::
        P[i]_d = \\sum_{j} (-\\omega^2)^{\\delta_{ij}} |j\\rangle\\langle j|

    where

    .. math::
        \\omega = \\exp(\\frac{2\\pi i}{d})

    and d is the number of levels (2 levels is a qubit,
    3 levels is a qutrit, etc.)

    References:
        - https://pubs.aip.org/aip/jmp/article-abstract/56/3/032202/763827
    """

    _num_qudits = 1
    _num_params = 0

    def __init__(self, index: int, radix: int = 3) -> None:
        """
        Construct a PDGate.

        Args:
            index (int): The level for the phase qudit gate (<radix).

            radix (int): The number of qudit levels (>=2). Defaults to
                qutrits.

        Raises:
            ValueError: if radix < 2

            ValueError: if index >= radix
        """
        if not is_integer(radix):
            raise TypeError(f'Expected integer for radix, got {type(radix)}.')

        if radix < 2:
            raise ValueError(f'Radix must be greater than 1, got {radix}.')

        if not is_integer(index):
            raise TypeError(f'Expected integer for index, got {type(index)}.')

        if index >= radix:
            raise ValueError(
                'PDGate index must be less than number of levels:'
                f'got: {index=} >= {radix=}.',
            )

        self._radix = radix
        self.index = index

        # Calculate unitary
        omega = np.exp(2j * np.pi * index / radix)
        diags = [(-omega ** 2) if i == index else 1 for i in range(radix)]
        self._utry = UnitaryMatrix(np.diag(diags), self.radixes)



"""This module implements the HGate."""
from __future__ import annotations

from math import pi
from math import sqrt

from numpy import array
from numpy import complex128
from numpy import exp
from numpy import zeros

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class HGate(ConstantGate, QuditGate):
    """
    The one-qudit Hadamard gate. This is a Clifford gate.

    By default, the HGate is a qubit Hadamard gate. However, it can be
    generalized to qudits by setting the radix parameter during construction.

    It is represented by the following matrix for qubits:

    .. math::

        \\begin{pmatrix}
        \\frac{\\sqrt{2}}{2} & \\frac{\\sqrt{2}}{2} \\\\
        \\frac{\\sqrt{2}}{2} & -\\frac{\\sqrt{2}}{2} \\\\
        \\end{pmatrix}

    However, generally it is given by the following formula:

    .. math::
        H = \\frac{1}{\\sqrt(d)} \\sum_{ij} \\omega_d^{ij} |i\\rangle\\langle j|

    where

    .. math::
        \\omega = \\exp(\\frac{2\\pi i}{d})

    and `d` is the number of levels (2 levels is a qubit,
    3 levels is a qutrit, etc.)

    References:
        - https://www.frontiersin.org/articles/10.3389/fphy.2020.589504/full
        - https://pubs.aip.org/aip/jmp/article-abstract/56/3/032202/763827
    """

    _num_qudits = 1
    _qasm_name = 'h'

    def __init__(self, radix: int = 2) -> None:
        """
        Construct a HGate.

        Args:
            radix (int): The number of qudit levels (>=2). By default, this
                is 2, which is gives a qubit Hadamard gate.

        Raises:
            ValueError: if radix < 2
        """
        if not is_integer(radix):
            raise TypeError(f'Expected integer for radix, got {type(radix)}.')

        if radix < 2:
            raise ValueError(f'Radix must be greater than 1, got {radix}.')

        self._radix = radix

        # Calculate unitary
        if radix == 2:
            matrix = array(
                [
                    [sqrt(2) / 2, sqrt(2) / 2],
                    [sqrt(2) / 2, -sqrt(2) / 2],
                ],
                dtype=complex128,
            )
            self._utry = UnitaryMatrix(matrix)

        else:
            matrix = zeros([radix] * 2, dtype=complex128)
            omega = exp(2j * pi / radix)
            for i in range(radix):
                for j in range(i, radix):
                    val = omega ** (i * j)
                    matrix[i, j] = val
                    matrix[j, i] = val
            matrix *= 1 / sqrt(radix)
            self._utry = UnitaryMatrix(matrix, self.radixes)



"""This module implements the QubitGate base class."""
from __future__ import annotations

from bqskit.ir.gate import Gate


class QubitGate(Gate):
    """A gate that only acts on qubits."""

    @property
    def radixes(self) -> tuple[int, ...]:
        """The number of orthogonal states for each qudit."""
        return tuple([2] * self.num_qudits)



"""This module implements the CircuitIterator class."""
from __future__ import annotations

import heapq
from typing import Iterator
from typing import Sequence
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

from bqskit.ir.operation import Operation
from bqskit.ir.point import CircuitPoint
from bqskit.ir.point import CircuitPointLike
from bqskit.ir.region import CircuitRegion
from bqskit.ir.region import CircuitRegionLike
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_sequence

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit


class CircuitIterator(Iterator[Union[Operation, Tuple[int, Operation]]]):
    """A CircuitIterator iterates through a circuit in a simulation order."""

    def __init__(
            self,
            circuit: Circuit,
            start: CircuitPointLike = CircuitPoint(0, 0),
            end: CircuitPointLike | None = None,
            qudits_or_region: CircuitRegionLike | Sequence[int] | None = None,
            exclude: bool = False,
            reverse: bool = False,
            and_cycles: bool = False,
    ) -> None:
        """
        Construct a CircuitIterator.

        Args:
            circuit (Circuit): The circuit to iterate through.

            start (CircuitPointLike): Only iterate through points greater
                than or equal to `start`. Defaults to start at the beginning
                of the circuit. (Default: (0, 0))

            end (CircuitPointLike | None): Only iterate through points
                less than or equal to this. If left as None, iterates
                until the end of the circuit. (Default: None)

            qudits_or_region (CircuitRegionLike | Sequence[int] | None):
                Determines the way the circuit is iterated. If a region
                is given, then iterate through operations in the region.
                If a sequence of qudit indices is given, then only iterate
                the operations touching those qudits. If left as None,
                then iterate through the entire circuit in simulation order.
                (Default: None)

            exclude (bool): If iterating through a region or only some
                qudits and `exclude` is true, then do not yield operations
                that are only partially in the region or on the desired
                qudits. This may result in a sequence of operations that
                does not occur in simulation order in the circuit.
                (Default: False)

            reverse (bool): Reverse the ordering. If true, then end acts
                as start and vice versa. (Default: False)

            and_cycles (bool): If true, in addition to the operation,
                return the cycle index where it was found. (Default: False)
        """
        if (
            start == (0, 0)
            and end is None
            and qudits_or_region is None
            and not exclude
            and not reverse
        ):
            # Default iteration can be done faster with the DAG Iterator
            self.iter: CircuitIterator = CircuitDagIterator(circuit, and_cycles)

        else:
            self.iter = CircuitGridIterator(
                circuit,
                start,
                end,
                qudits_or_region,
                exclude,
                reverse,
                and_cycles,
            )

    def __next__(self) -> Operation | tuple[int, Operation]:
        return self.iter.__next__()

    def __iter__(self) -> CircuitIterator:
        return self.iter.__iter__()


class CircuitDagIterator(CircuitIterator):
    """Fast and simple iteration through circuit."""

    def __init__(self, circuit: Circuit, and_cycles: bool = False):
        """Set a CircuitDagIterator to iterate through `circuit`."""
        self.circuit = circuit
        self.and_cycles = and_cycles

        # Track frontier for topological traversal
        self.frontier = list(circuit.front)
        heapq.heapify(self.frontier)

        # Tracks how many of a gate's dependencies have been partitioned
        self.prev_binned_counts = {n: 0 for n in circuit.front}

    def __next__(self) -> Operation | tuple[int, Operation]:
        if len(self.frontier) == 0:
            raise StopIteration

        point = heapq.heappop(self.frontier)
        self.prev_binned_counts.pop(point)

        # Update frontier
        for successor in self.circuit.next(point):
            if successor not in self.prev_binned_counts:
                self.prev_binned_counts[successor] = 1
            else:
                self.prev_binned_counts[successor] += 1
            num_prev_binned = self.prev_binned_counts[successor]
            total_num_prev = len(self.circuit.prev(successor))
            if num_prev_binned == total_num_prev:
                heapq.heappush(self.frontier, successor)

        op = self.circuit[point]
        return (point.cycle, op) if self.and_cycles else op

    def __iter__(self) -> CircuitIterator:
        return self


class CircuitGridIterator(CircuitIterator):
    """
    The CircuitGridIterator Class.

    A CircuitIterator can iterate through a circuit in a few different ways. By
    default it can iterate through all operations in the circuit in simulation
    order. Additionally, it can iterate all the operations on a qudit or set of
    qudits or iterate through a specified CircuitRegion. If iterating all
    operations in a region or on some qudits, you can choose to exclude
    operations that only are partially in the specified area.
    """

    def __init__(
            self,
            circuit: Circuit,
            start: CircuitPointLike = CircuitPoint(0, 0),
            end: CircuitPointLike | None = None,
            qudits_or_region: CircuitRegionLike | Sequence[int] | None = None,
            exclude: bool = False,
            reverse: bool = False,
            and_cycles: bool = False,
    ) -> None:
        """
        Construct a CircuitGridIterator.

        Args:
            circuit (Circuit): The circuit to iterate through.

            start (CircuitPointLike): Only iterate through points greater
                than or equal to `start`. Defaults to start at the beginning
                of the circuit. (Default: (0, 0))

            end (CircuitPointLike | None): Only iterate through points
                less than or equal to this. If left as None, iterates
                until the end of the circuit. (Default: None)

            qudits_or_region (CircuitRegionLike | Sequence[int] | None):
                Determines the way the circuit is iterated. If a region
                is given, then iterate through operations in the region.
                If a sequence of qudit indices is given, then only iterate
                the operations touching those qudits. If left as None,
                then iterate through the entire circuit in simulation order.
                (Default: None)

            exclude (bool): If iterating through a region or only some
                qudits and `exclude` is true, then do not yield operations
                that are only partially in the region or on the desired
                qudits. This may result in a sequence of operations that
                does not occur in simulation order in the circuit.
                (Default: False)

            reverse (bool): Reverse the ordering. If true, then end acts
                as start and vice versa. (Default: False)

            and_cycles (bool): If true, in addition to the operation,
                return the cycle index where it was found. (Default: False)
        """
        if not CircuitPoint.is_point(start):
            raise TypeError(f'Expected point for start, got {type(start)}.')

        if end is not None and not CircuitPoint.is_point(end):
            raise TypeError(f'Expected point for end, got {type(end)}.')

        if end is None:
            end = CircuitPoint(
                circuit.num_cycles - 1,
                circuit.num_qudits - 1,
            )

        self.circuit = circuit
        self.start = CircuitPoint(*start)
        self.end = CircuitPoint(*end)
        self.exclude = exclude
        self.reverse = reverse
        self.and_cycles = and_cycles

        # Set mode of iteration:
        if qudits_or_region is None:
            # iterate through the entire circuit normally
            self.qudits = list(range(self.circuit.num_qudits))
            self.region = CircuitRegion({
                qudit: (0, self.circuit.num_cycles)
                for qudit in self.qudits
            })

        elif CircuitRegion.is_region(qudits_or_region):
            # iterate through the region in the circuit
            self.qudits = list(qudits_or_region.keys())
            self.region = CircuitRegion(qudits_or_region)

        elif is_sequence(qudits_or_region):
            # iterate through the circuit but only on the specified qudits
            if not all(is_integer(qudit) for qudit in qudits_or_region):
                raise TypeError('Expected region or sequence of indices.')

            if not all(
                0 <= qudit < self.circuit.num_qudits
                for qudit in qudits_or_region
            ):
                raise ValueError('Invalid sequence of qudit indices.')

            self.qudits = list(qudits_or_region)
            self.region = CircuitRegion({
                qudit: (0, self.circuit.num_cycles)
                for qudit in self.qudits
            })

        self.max_qudit = max(self.qudits)
        self.min_qudit = min(self.qudits)
        self.min_cycle = self.region.min_cycle
        self.max_cycle = self.region.max_cycle

        if self.start < (self.min_cycle, self.min_qudit):
            self.start = CircuitPoint(self.min_cycle, self.min_qudit)

        if self.end > (self.max_cycle, self.max_qudit):
            self.end = CircuitPoint(self.max_cycle, self.max_qudit)

        assert isinstance(self.start, CircuitPoint)  # TODO: Typeguard
        assert isinstance(self.end, CircuitPoint)  # TODO: Typeguard

        # Pointer into the circuit structure
        self.cycle = self.start.cycle if not self.reverse else self.end.cycle
        self.qudit = self.start.qudit if not self.reverse else self.end.qudit

        # Used to track changes to circuit structure
        self.num_ops = self.circuit.num_operations
        self.num_cycles = self.circuit.num_cycles
        self.num_qudits = self.circuit.num_qudits

        # Ensure operations are only returned once
        self.qudits_to_skip: set[int] = set()

    def increment_iter(self) -> None:
        """Increment the iterator to the next valid circuit point."""
        while (
            self.qudit in self.qudits_to_skip
            or self.qudit not in self.qudits
            or (
                self.cycle not in self.region[self.qudit]
                and self.cycle <= self.max_cycle
            )
        ):
            self.qudit += 1

            if self.qudit > self.max_qudit:
                self.qudit = self.min_qudit
                self.cycle += 1
                self.qudits_to_skip.clear()

    def decrement_iter(self) -> None:
        """Decrement the iterator to the next valid circuit point."""
        while (
            self.qudit in self.qudits_to_skip
            or self.qudit not in self.qudits
            or (
                self.cycle not in self.region[self.qudit]
                and self.cycle >= self.min_cycle
            )
        ):
            self.qudit -= 1

            if self.qudit < self.min_qudit:
                self.qudit = self.max_qudit
                self.cycle -= 1
                self.qudits_to_skip.clear()

    def step(self) -> None:
        """Move the iterator one step."""
        if not self.reverse:
            self.increment_iter()
        else:
            self.decrement_iter()

        point = (self.cycle, self.qudit)
        if point < self.start or point > self.end:
            raise StopIteration

    def __next__(self) -> Operation | tuple[int, Operation]:
        if (
            self.num_ops != self.circuit.num_operations
            or self.num_cycles != self.circuit.num_cycles
            or self.num_qudits != self.circuit.num_qudits
        ):
            raise RuntimeError('Circuit changed under iteration.')

        while True:
            self.step()
            op = self.circuit._circuit[self.cycle][self.qudit]

            if op is None:
                self.qudits_to_skip.add(self.qudit)
                continue

            self.qudits_to_skip.update(op.location)

            if self.exclude:
                if not all(qudit in self.qudits for qudit in op.location):
                    continue

                if not all(
                    self.region.overlaps((self.cycle, qudit))
                    for qudit in op.location
                ):
                    continue

            if self.and_cycles:
                return self.cycle, op

            return op

    def __iter__(self) -> CircuitIterator:
        return self



"""Register MachineModel specific default workflows."""
from __future__ import annotations

import warnings

from bqskit.compiler.machine import MachineModel
from bqskit.compiler.workflow import Workflow
from bqskit.compiler.workflow import WorkflowLike


_compile_circuit_registry: dict[MachineModel, dict[int, Workflow]] = {}
_compile_unitary_registry: dict[MachineModel, dict[int, Workflow]] = {}
_compile_stateprep_registry: dict[MachineModel, dict[int, Workflow]] = {}
_compile_statemap_registry: dict[MachineModel, dict[int, Workflow]] = {}


def model_registered_target_types(key: MachineModel) -> list[str]:
    """
    Return a list of target_types for which key is registered.

    Args:
        key (MachineModel): A MachineModel to check for.

    Returns:
        (list[str]): If `key` has been registered in any of the registry, the
            name of that target type will be contained in this list.
    """
    registered_types = []
    if key in _compile_circuit_registry:
        registered_types.append('circuit')
    if key in _compile_unitary_registry:
        registered_types.append('unitary')
    if key in _compile_stateprep_registry:
        registered_types.append('stateprep')
    if key in _compile_statemap_registry:
        registered_types.append('statemap')
    return registered_types


def register_workflow(
    key: MachineModel,
    workflow: WorkflowLike,
    optimization_level: int,
    target_type: str = 'circuit',
) -> None:
    """
    Register a workflow for a given MachineModel.

    The _compile_registry enables MachineModel specific workflows to be
    registered for use in the `bqskit.compile` method. _compile_registry maps
    MachineModels a dictionary of Workflows which are indexed by optimization
    level. This object should not be accessed directly by the user, but
    instead through the `register_workflow` function.

    Args:
        key (MachineModel): A MachineModel to register the workflow under.
            If a circuit is compiled targeting this machine or gate set, the
            registered workflow will be used.

        workflow (list[BasePass]): The workflow or list of passes that will
            be executed if the MachineModel in a call to `compile` matches
            `key`. If `key` is already registered, a warning will be logged.

        optimization_level (int): The optimization level with which to
            register the workflow.

        target_type (str): Register a workflow for targets of this type. Must
            be 'circuit', 'unitary', 'stateprep', or 'statemap'.
            (Default: 'circuit')

    Example:
        model_t = SpecificMachineModel(num_qudits, radixes)
        workflow = [QuickPartitioner(3), NewFangledOptimization()]
        register_workflow(model_t, workflow, level, 'circuit')
        ...
        new_circuit = compile(circuit, model_t, optimization_level=level)

    Raises:
        Warning: If a workflow for a given optimization_level is overwritten.

        ValueError: If `target_type` is not 'circuit', 'unitary', 'stateprep',
            or 'statemap'.
    """
    if target_type not in ['circuit', 'unitary', 'stateprep', 'statemap']:
        m = 'target_type must be "circuit", "unitary", "stateprep", or '
        m += f'"statemap", got {target_type}.'
        raise ValueError(m)

    if target_type == 'circuit':
        _compile_registry = _compile_circuit_registry
    elif target_type == 'unitary':
        _compile_registry = _compile_unitary_registry
    elif target_type == 'stateprep':
        _compile_registry = _compile_stateprep_registry
    else:
        _compile_registry = _compile_statemap_registry

    workflow = Workflow(workflow)

    new_workflow = {optimization_level: workflow}
    if key in _compile_registry:
        if optimization_level in _compile_registry[key]:
            m = f'Overwritting workflow for {key} at level '
            m += f'{optimization_level} for target type {target_type}.'
            m += 'If multiple Namespace packages are installed, ensure'
            m += 'that their __init__.py files do not attempt to'
            m += 'overwrite the same default Workflows.'
            warnings.warn(m)
        _compile_registry[key].update(new_workflow)
    else:
        _compile_registry[key] = new_workflow



"""This module implements the DijkstraHeuristic class."""
from __future__ import annotations

from bqskit.ir.circuit import Circuit
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


# TODO: Should name be changed to Breadth
class DijkstraHeuristic(HeuristicFunction):
    """
    The DijkstraHeuristic HeuristicFunction class.

    Defines a heuristic that relies only on circuit depth, which guarantees a
    minimal-depth final solution at the expense of a long runtime. This will
    create a behavior similar to breadth-first search.
    """

    def get_value(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> float:
        """Return the heuristic's value, see HeuristicFunction for more info."""
        cost = 0.0
        for gate in circuit.gate_set:
            if gate.num_qudits == 1:
                continue
            cost += float(circuit.count(gate))
        return cost



"""This module implements the IdentityGate."""
from __future__ import annotations

from typing import Sequence

import numpy as np

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_valid_radixes


class IdentityGate(ConstantGate):
    """An Identity (No-OP) Gate."""

    def __init__(
        self,
        num_qudits: int = 1,
        radixes: Sequence[int] = [],
    ) -> None:
        """
        Create an IdentityGate, defaulting to a single-qubit identity.

        Args:
            num_qudits (int): The number of qudits this gate acts on.

            radixes (Sequence[int]): The number of orthogonal
                states for each qudit. Defaults to qubits.

        Raises:
            ValueError: If `num_qudits` is nonpositive.
        """
        if num_qudits <= 0:
            raise ValueError('Expected positive integer, got %d' % num_qudits)

        if len(radixes) != 0 and not is_valid_radixes(radixes, num_qudits):
            raise TypeError('Invalid radixes.')

        self._num_qudits = num_qudits
        self._radixes = tuple(radixes or [2] * num_qudits)
        self._dim = int(np.prod(self.radixes))
        self._utry = UnitaryMatrix.identity(self.dim, self.radixes)
        self._qasm_name = 'identity%d' % self.num_qudits

    def get_qasm_gate_def(self) -> str:
        """Return a qasm gate definition block for this gate."""
        param_symbols = ['a%d' % i for i in range(self.num_qudits)]
        param_str = ','.join(param_symbols)
        header = 'gate identity%d %s' % (self.num_qudits, param_str)
        body_stmts = ['\tU(0,0,0) %s;' % sym for sym in param_symbols]
        body = '\n'.join(body_stmts)
        return f'{header}\n{{\n{body}\n}}\n'

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, IdentityGate)
            and self.num_qudits == other.num_qudits
            and self.radixes == other.radixes
        )

    def __hash__(self) -> int:
        return hash((self.num_qudits, self.radixes))

    def __str__(self) -> str:
        if self.is_qubit_only():
            return f'IdentityGate({self.num_qudits})'
        else:
            return f'IdentityGate({self.num_qudits}, {self.radixes})'



"""This module implements the CircuitGate class."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.ir.location import CircuitLocation
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit

from bqskit.ir.gate import Gate


class CircuitGate(Gate):
    """
    The CircuitGate class.

    A CircuitGate is a immutable circuit represented as a gate.
    """

    def __init__(self, circuit: Circuit, move: bool = False) -> None:
        """
        CircuitGate Constructor.

        Args:
            circuit (Circuit): The circuit to copy into gate format.

            move (bool): If true, the constructor will not copy the circuit.
                This should only be used when you are sure `circuit` will no
                longer be used on caller side. If unsure use the default.
                (Default: False)
        """

        self._circuit = circuit if move else circuit.copy()
        self._num_qudits = self._circuit.num_qudits
        self._radixes = self._circuit.radixes
        self._num_params = self._circuit.num_params
        self._name = 'CircuitGate(%s)' % str(self._circuit)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        return self._circuit.get_unitary(params)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        return self._circuit.get_grad(params)

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        return self._circuit.get_unitary_and_grad(params)

    def is_differentiable(self) -> bool:
        """Return true if the circuit is differentiable."""
        return self._circuit.is_differentiable()

    def __hash__(self) -> int:
        hashes: list[int] = [hash(self.name)]
        for op in self._circuit:
            hashes.append(hash(op))

            # Don't let the hash list grow too large.
            if len(hashes) >= 100:
                hashes = [hash(tuple(hashes))]

        hash_val = hash(tuple(hashes)) if len(hashes) > 1 else hashes[0]
        return hash_val

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CircuitGate):
            return NotImplemented

        if self._circuit.num_qudits != other._circuit.num_qudits:
            return False

        if self._circuit.radixes != other._circuit.radixes:
            return False

        return all(
            op1.gate == op2.gate and op1.location == op2.location
            for op1, op2 in zip(self._circuit, other._circuit)
        )

    def get_qasm_gate_def(self) -> str:
        """Returns a qasm gate definition block for this gate."""
        ret = ''
        for gate in self._circuit.gate_set:
            ret += gate.get_qasm_gate_def()

        # TODO: replace use of hash
        id = hash(self)
        if id < 0:
            id = -id
        ret += f'gate circuitgate_{id} '

        if self.num_params > 0:
            ret += '(p'
            ret += ', p'.join([str(i) for i in range(self.num_params)])
            ret += ') '

        ret += 'q' + ', q'.join([str(i) for i in range(self.num_qudits)])

        ret += ' {\n'
        param_index = 0
        for op in self._circuit:
            params = [
                f'p{i}'
                for i in range(param_index, param_index + op.num_params)
            ]
            if isinstance(op.gate, CircuitGate):
                op_id = hash(op.gate)
                if op_id < 0:
                    op_id = -op_id
                ret += '\t{}({}) q{};\n'.format(
                    f'circuitgate_{op_id}',
                    ', '.join([str(p) for p in params]),
                    ', q'.join([str(q) for q in op.location]),
                ).replace('()', '')
            else:
                ret += '\t{}({}) q{};\n'.format(
                    op.gate.qasm_name,
                    ', '.join([str(p) for p in params]),
                    ', q'.join([str(q) for q in op.location]),
                ).replace('()', '')
            param_index += op.num_params
        ret += '}\n'
        return ret

    def get_qasm(self, params: RealVector, location: CircuitLocation) -> str:
        """Returns the qasm string for this gate."""
        # TODO: replace use of hash
        id = hash(self)
        if id < 0:
            id = -id

        return '{}({}) q[{}];\n'.format(
            f'circuitgate_{id}',
            ', '.join([str(p) for p in params]),
            '], q['.join([str(q) for q in location]),
        ).replace('()', '')



"""This module implements the UnitaryBuilder class."""
from __future__ import annotations

import logging
from typing import cast
from typing import Sequence
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitary import Unitary
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_valid_radixes

if TYPE_CHECKING:
    from bqskit.ir.location import CircuitLocationLike

logger = logging.getLogger(__name__)


class UnitaryBuilder(Unitary):
    """
    An object for fast unitary accumulation using tensor networks.

    A UnitaryBuilder is similar to a StringBuilder in the sense that it is an
    efficient way to string together or accumulate :class:`Unitary` objects.
    This class uses concepts from tensor networks to efficiently multiply
    unitary matrices.
    """

    def __init__(self, num_qudits: int, radixes: Sequence[int] = []) -> None:
        """
        UnitaryBuilder constructor.

        Args:
            num_qudits (int): The number of qudits to build a Unitary for.

            radixes (Sequence[int]): A sequence with its length equal
                to `num_qudits`. Each element specifies the base of a
                qudit. Defaults to qubits.

        Raises:
            ValueError: If `num_qudits` is nonpositive.

            ValueError: If the length of `radixes` is not equal to
                `num_qudits`.

        Examples:
            >>> builder = UnitaryBuilder(4)  # Creates a 4-qubit builder.
        """

        if not is_integer(num_qudits):
            raise TypeError(
                'Expected int for num_qudits, got %s.' %
                type(num_qudits),
            )

        if num_qudits <= 0:
            raise ValueError(
                'Expected positive number for num_qudits, got %d.' %
                num_qudits,
            )

        self._num_qudits = num_qudits
        self._radixes = tuple(radixes if len(radixes) > 0 else [2] * num_qudits)

        if not is_valid_radixes(self.radixes):
            raise TypeError('Invalid qudit radixes.')

        if len(self.radixes) != self.num_qudits:
            raise ValueError(
                'Expected length of radixes to be equal to num_qudits:'
                ' %d != %d' % (len(self.radixes), self.num_qudits),
            )

        self._num_params = 0
        self._dim = int(np.prod(self.radixes))
        self.tensor = np.identity(self.dim, dtype=np.complex128)
        self.tensor = self.tensor.reshape(self.radixes * 2)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Build the unitary, see :func:`Unitary.get_unitary` for more."""
        utry = self.tensor.reshape((self.dim, self.dim))
        return UnitaryMatrix(utry, self.radixes, False)

    def apply_right(
        self,
        utry: UnitaryMatrix,
        location: CircuitLocationLike,
        inverse: bool = False,
        check_arguments: bool = True,
    ) -> None:
        """
        Apply the specified unitary on the right of this UnitaryBuilder.

        ..
                 .-----.   .------.
              0 -|     |---|      |-
              1 -|     |---| utry |-
                 .     .   '------'
                 .     .
                 .     .
            n-1 -|     |------------
                 '-----'

        Args:
            utry (UnitaryMatrix): The unitary to apply.

            location (CircuitLocationLike): The qudits to apply the unitary on.

            inverse (bool): If true, apply the inverse of the unitary.

            check_arguments (bool): If true, check the inputs for type and
                value errors.

        Raises:
            ValueError: If `utry`'s size does not match the given location.

            ValueError: if `utry`'s radixes does not match the given location.

        Notes:
            - Applying the unitary on the right is equivalent to multiplying
              the unitary on the left of the tensor. The notation comes
              from the quantum circuit perspective.

            - This operation is performed using tensor contraction.
        """
        from bqskit.ir.location import CircuitLocation
        if check_arguments:
            if not isinstance(utry, UnitaryMatrix):
                raise TypeError('Expected UnitaryMatrix, got %s', type(utry))

            if not CircuitLocation.is_location(location, self.num_qudits):
                raise TypeError('Invalid location.')

            location = CircuitLocation(location)

            if len(location) != utry.num_qudits:
                raise ValueError('Unitary and location size mismatch.')

            for utry_radix, bldr_radix_idx in zip(utry.radixes, location):
                if utry_radix != self.radixes[bldr_radix_idx]:
                    raise ValueError('Unitary and location radix mismatch.')

        left_perm = list(cast(CircuitLocation, location))
        mid_perm = [x for x in range(self.num_qudits) if x not in left_perm]
        right_perm = [x + self.num_qudits for x in range(self.num_qudits)]

        left_dim = int(np.prod([self.radixes[x] for x in left_perm]))

        utry = utry.dagger if inverse else utry

        perm = left_perm + mid_perm + right_perm
        self.tensor = self.tensor.transpose(perm)
        self.tensor = self.tensor.reshape((left_dim, -1))
        self.tensor = utry @ self.tensor

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        self.tensor = self.tensor.reshape(shape)
        inv_perm = list(np.argsort(perm))
        self.tensor = self.tensor.transpose(inv_perm)

    def apply_left(
        self,
        utry: UnitaryMatrix,
        location: CircuitLocationLike,
        inverse: bool = False,
        check_arguments: bool = True,
    ) -> None:
        """
        Apply the specified unitary on the left of this UnitaryBuilder.

        ..
                 .------.   .-----.
              0 -|      |---|     |-
              1 -| gate |---|     |-
                 '------'   .     .
                            .     .
                            .     .
            n-1 ------------|     |-
                            '-----'

        Args:
            utry (UnitaryMatrix): The unitary to apply.

            location (CircuitLocationLike): The qudits to apply the unitary on.

            inverse (bool): If true, apply the inverse of the unitary.

            check_arguments (bool): If true, check the inputs for type and
                value errors.

        Raises:
            ValueError: If `utry`'s size does not match the given location.

            ValueError: if `utry`'s radixes does not match the given location.

        Notes:
            - Applying the unitary on the left is equivalent to multiplying
              the unitary on the right of the tensor. The notation comes
              from the quantum circuit perspective.

            - This operation is performed using tensor contraction.
        """
        from bqskit.ir.location import CircuitLocation
        if check_arguments:
            if not isinstance(utry, UnitaryMatrix):
                raise TypeError('Expected UnitaryMatrix, got %s', type(utry))

            if not CircuitLocation.is_location(location, self.num_qudits):
                raise TypeError('Invalid location.')

            location = CircuitLocation(location)

            if len(location) != utry.num_qudits:
                raise ValueError('Unitary and location size mismatch.')

            for utry_radix, bldr_radix_idx in zip(utry.radixes, location):
                if utry_radix != self.radixes[bldr_radix_idx]:
                    raise ValueError('Unitary and location radix mismatch.')

        location = cast(CircuitLocation, location)
        left_perm = list(range(self.num_qudits))
        mid_perm = [
            x + self.num_qudits
            for x in left_perm
            if x not in location
        ]
        right_perm = [x + self.num_qudits for x in location]

        right_dim = int(
            np.prod([
                self.radixes[x - self.num_qudits]
                for x in right_perm
            ]),
        )

        utry = utry.dagger if inverse else utry

        perm = left_perm + mid_perm + right_perm
        self.tensor = self.tensor.transpose(perm)
        self.tensor = self.tensor.reshape((-1, right_dim))
        self.tensor = self.tensor @ utry

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        self.tensor = self.tensor.reshape(shape)
        inv_perm = list(np.argsort(perm))
        self.tensor = self.tensor.transpose(inv_perm)

    def eval_apply_right(
        self,
        M: npt.NDArray[np.complex128],
        location: CircuitLocationLike,
    ) -> npt.NDArray[np.complex128]:
        """
        Evaluate the application of `M` on the right of this UnitaryBuilder.

        See :func:`apply_right` for more info.
        """
        from bqskit.ir.location import CircuitLocation
        left_perm = list(cast(CircuitLocation, location))
        mid_perm = [x for x in range(self.num_qudits) if x not in left_perm]
        right_perm = [x + self.num_qudits for x in range(self.num_qudits)]

        left_dim = int(np.prod([self.radixes[x] for x in left_perm]))

        perm = left_perm + mid_perm + right_perm
        tensor_copy = self.tensor.copy()
        tensor_copy = tensor_copy.transpose(perm)
        tensor_copy = tensor_copy.reshape((left_dim, -1))
        tensor_copy = M @ tensor_copy  # TODO: Require out matrix to avoid copy

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        tensor_copy = tensor_copy.reshape(shape)
        inv_perm = list(np.argsort(perm))
        tensor_copy = tensor_copy.transpose(inv_perm)
        out_M = tensor_copy.reshape((self.dim, self.dim))
        return out_M

    def eval_apply_left(
        self,
        M: npt.NDArray[np.complex128],
        location: CircuitLocationLike,
    ) -> npt.NDArray[np.complex128]:
        """
        Evaluate the application of `M` on the left of this UnitaryBuilder.

        See :func:`apply_left` for more info.
        """
        from bqskit.ir.location import CircuitLocation
        location = cast(CircuitLocation, location)
        left_perm = list(range(self.num_qudits))
        mid_perm = [
            x + self.num_qudits
            for x in left_perm
            if x not in location
        ]
        right_perm = [x + self.num_qudits for x in location]

        right_dim = int(
            np.prod([
                self.radixes[x - self.num_qudits]
                for x in right_perm
            ]),
        )

        perm = left_perm + mid_perm + right_perm
        tensor_copy = self.tensor.copy()
        tensor_copy = tensor_copy.transpose(perm)
        tensor_copy = tensor_copy.reshape((-1, right_dim))
        tensor_copy = tensor_copy @ M

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        tensor_copy = tensor_copy.reshape(shape)
        inv_perm = list(np.argsort(perm))
        tensor_copy = tensor_copy.transpose(inv_perm)
        out_M = tensor_copy.reshape((self.dim, self.dim))
        return out_M

    def calc_env_matrix(
            self, location: Sequence[int],
    ) -> npt.NDArray[np.complex128]:
        """
        Calculates the environment matrix w.r.t. the specified location.

        Args:
            location (Sequence[int]): Calculate the environment matrix with
                respect to the qudit indices in location.

        Returns:
            np.ndarray: The environmental matrix.
        """

        left_perm = list(range(self.num_qudits))
        left_perm = [x for x in left_perm if x not in location]
        left_perm = left_perm + [x + self.num_qudits for x in left_perm]
        right_perm = list(location) + [x + self.num_qudits for x in location]

        perm = left_perm + right_perm
        a = np.transpose(self.tensor, perm)
        a = np.reshape(
            a, (
                2 ** (self.num_qudits - len(location)),
                2 ** (self.num_qudits - len(location)),
                2 ** len(location),
                2 ** len(location),
            ),
        )
        return np.trace(a)



"""This module implements the ToVariablePass."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.ir.point import CircuitPoint

_logger = logging.getLogger(__name__)


class ToVariablePass(BasePass):
    """Converts single-qudit general unitary gates to Variable Unitary Gates."""

    def __init__(self, convert_all_single_qudit_gates: bool = False) -> None:
        """
        Construct a ToVariablePass.

        Args:
            convert_all_single_qudit_gates (bool): Indicates wheter to convert
            only the general gates, or every single qudit gate.
        """
        self.convert_all_single_qudit_gates = convert_all_single_qudit_gates

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.debug(
            'Converting single-qudit general gates to VariableUnitaryGate.',
        )
        for cycle, op in circuit.operations_with_cycles():
            if (
                op.gate.num_qudits == 1 and (
                    isinstance(op.gate, GeneralGate)
                    or self.convert_all_single_qudit_gates
                )
            ):
                params = VariableUnitaryGate.get_params(op.get_unitary())
                point = CircuitPoint(cycle, op.location[0])
                vgate = VariableUnitaryGate(op.num_qudits, op.radixes)
                circuit.replace_gate(
                    point, vgate, op.location, params,
                )



"""This module implements the DiscreteLayerGenerator class."""
from __future__ import annotations

import logging
from typing import Callable
from typing import Sequence

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.gates.constant.h import HGate
from bqskit.ir.gates.constant.t import TGate
from bqskit.ir.gates.parameterized.pauliz import PauliZGate
from bqskit.ir.operation import Operation
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_sequence


_logger = logging.getLogger(__name__)


class DiscreteLayerGenerator(LayerGenerator):
    """
    The DiscreteLayerGenerator class.

    Expands circuits using only discrete gates. This is a non-reinforcement
    learning version of diagonalizing in
    https://arxiv.org/abs/2409.00433.
    """

    def __init__(
        self,
        gateset: Sequence[Gate] = [HGate(), TGate(), CNOTGate()],
        double_headed: bool = False,
        dividing_gate_type: Callable[[int], Gate] = PauliZGate,
    ) -> None:
        """
        Construct a DiscreteLayerGenerator.

        Args:
            gateset (Sequence[Gate]): A sequence of gates that can be used
                in the output circuit. These must be non-parameterized gates.
                (Default: [HGate, TGate, CNOTGate])

            double_headed (bool): If True, successors will be generated by
                both appending and prepending gates. This lets unitaries be
                diagonalized instead of inverted. (Default: False)

            dividing_gate (Callable[[int], Gate]): A gate that goes between
                the two heads of the discrete searches. If double_headed is
                False, this gate simply goes at the beggining of the circuit.
                (Default: PauliZGate)

        Raises:
            ValueError: If the gateset is not a sequence.

            ValueError: If the gateset contains a parameterized gate.

            ValueError: If the radices of gates are different.

        TODO:
            Check universality of gateset.
        """
        if not is_sequence(gateset):
            m = f'Expected sequence of gates, got {type(gateset)}.'
            raise ValueError(m)

        radix = gateset[0].radixes[0]
        for gate in gateset:
            if gate.num_params > 0:
                m = 'Expected gate for constant gates, got parameterized'
                m += f' {gate} gate.'
                raise ValueError(m)
            for rad in gate.radixes:
                if rad != radix:
                    m = f'Radix mismatch on gate: {gate}. '
                    m += f'Expected {radix}, got {rad}.'
                    raise ValueError(m)
        self.gateset = gateset
        self.double_headed = double_headed
        self.dividing_gate_type = dividing_gate_type

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a radix mismatch with
                `self.initial_layer_gate`.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            m = f'Expected unitary or state, got {type(target)}.'
            raise TypeError(m)

        for radix in target.radixes:
            if radix != self.gateset[0].radixes[0]:
                m = 'Radix mismatch between target and gateset.'
                raise ValueError(m)

        init_circuit = Circuit(target.num_qudits, target.radixes)

        if self.double_headed:
            n = target.num_qudits
            span = list(range(n))
            init_circuit.append_gate(self.dividing_gate_type(n), span)

        return init_circuit

    def cancels_something(
        self,
        circuit: Circuit,
        gate: Gate,
        location: tuple[int, ...],
    ) -> bool:
        """Ensure applying gate at location does not cancel a previous gate."""
        last_cycle = circuit.num_cycles - 1
        try:
            op = circuit.get_operation((last_cycle, location[0]))
            op_gate, op_location = op.gate, op.location
            if op_location == location and op_gate.get_inverse() == gate:
                return True
            return False
        except IndexError:
            return False

    def count_repeats(
        self,
        circuit: Circuit,
        gate: Gate,
        qudit: int,
    ) -> int:
        """Count the number of times the last gate is repeated on qudit."""
        count = 0
        for cycle in reversed(range(circuit.num_cycles)):
            try:
                op = circuit.get_operation((cycle, qudit))
                if op.gate == gate:
                    count += 1
                else:
                    return count
            except IndexError:
                continue
        return count

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """
        if not isinstance(circuit, Circuit):
            raise TypeError(f'Expected circuit, got {type(circuit)}.')

        if circuit.num_qudits < 2:
            raise ValueError('Cannot expand a single-qudit circuit.')

        # Get the coupling graph
        coupling_graph = data.connectivity

        # Generate successors
        successors = []
        hashes = set()
        singles = [gate for gate in self.gateset if gate.num_qudits == 1]
        multis = [gate for gate in self.gateset if gate.num_qudits > 1]

        def add_to_successors(circuit: Circuit) -> None:
            h = self.hash_circuit_structure(circuit)
            if h not in hashes:
                successors.append(circuit)
                hashes.add(h)

        for gate in singles:
            for qudit in range(circuit.num_qudits):
                if gate.radixes[0] != circuit.radixes[qudit]:
                    continue
                if self.cancels_something(circuit, gate, (qudit,)):
                    continue
                if isinstance(gate, TGate):
                    if self.count_repeats(circuit, TGate(), qudit) >= 7:
                        continue
                successor = circuit.copy()
                successor.append_gate(gate, [qudit])

                add_to_successors(successor)

                if self.double_headed:
                    successor = circuit.copy()
                    op = Operation(gate, [qudit])
                    successor.insert(0, op)
                    add_to_successors(successor)

        for gate in multis:
            for edge in coupling_graph:
                if self.cancels_something(circuit, gate, edge):
                    continue
                qudit_radixes = [circuit.radixes[q] for q in edge]
                if gate.radixes != qudit_radixes:
                    continue
                successor = circuit.copy()
                successor.append_gate(gate, edge)
                add_to_successors(successor)

                if self.double_headed:
                    successor = circuit.copy()
                    op = Operation(gate, edge)
                    successor.insert(0, op)
                    add_to_successors(successor)

        return successors

    def hash_circuit_structure(self, circuit: Circuit) -> int:
        hashes = []
        for op in circuit:
            hashes.append(hash(op))
        return hash(tuple(hashes))



"""This module implements the Minimizer base class."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.qis.unitary.unitary import RealVector

if TYPE_CHECKING:
    from bqskit.ir.opt.cost.function import CostFunction


class Minimizer(abc.ABC):
    """
    The Minimizer class.

    An minimizer finds the parameters for a circuit template that minimizes some
    CostFunction.
    """

    @abc.abstractmethod
    def minimize(
        self, cost: CostFunction,
        x0: RealVector,
    ) -> npt.NDArray[np.float64]:
        """
        Minimize `cost` starting from the initial point `x0`.

        Args:
            cost (CostFunction): The CostFunction to minimize.

            x0 (np.ndarray): The initial point.

        Returns:
            (np.ndarray): The inputs that best minimizes the cost.

        Notes:
            This function should be side-effect free. This is because many
            calls may be running in parallel.
        """



"""
==========================================================
BQSKit Utility Classes and Functions (:mod:`bqskit.utils`)
==========================================================

.. currentmodule:: bqskit.utils

.. rubric:: Utility Classes

.. autosummary::
    :toctree: autogen
    :recursive:

    CachedClass

.. rubric:: Utility Modules

.. autosummary::
    :toctree: autogen
    :recursive:

    math
    typing
    random
    docs
    test.strategies
    test.types
"""
from __future__ import annotations

from bqskit.utils.cachedclass import CachedClass

__all__ = ['CachedClass']



"""This module implements the ExtractMeasurements and RestoreMeasurements
passes."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.measure import MeasurementPlaceholder
from bqskit.ir.point import CircuitPoint
_logger = logging.getLogger(__name__)


class ExtractMeasurements(BasePass):
    """A pass that stores and removes the measurements in a circuit."""

    key = '__measurement_data__'

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.info('Extracting measurements from circuit.')
        points_to_remove: list[CircuitPoint] = []
        cregs: dict[str, int] = {}
        measurements: dict[int, tuple[str, int]] = {}
        for cycle, op in circuit.operations_with_cycles():
            if isinstance(op.gate, MeasurementPlaceholder):
                points_to_remove.append(CircuitPoint(cycle, op.location[0]))
                measurements.update(op.gate.measurements)
                for name, size in op.gate.classical_regs:
                    if name not in cregs:
                        cregs[name] = size

        if len(cregs) > 0:
            circuit.batch_pop(points_to_remove)
            data[self.key] = (cregs, measurements)
            _logger.debug(f'Extracted classical registers: {cregs}.')
            _logger.debug(f'Extracted measurements: {measurements}.')


class RestoreMeasurements(BasePass):
    """A pass that restores the measurements in a circuit."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.info('Restoring measurements to circuit.')
        if ExtractMeasurements.key in data:
            cregs, measurements = data[ExtractMeasurements.key]

            # permute for final layout
            if 'final_mapping' in data:
                pi = data['final_mapping']
                measurements = {pi[q]: c for q, c in measurements.items()}

            mph = MeasurementPlaceholder(list(cregs.items()), measurements)
            circuit.append_gate(mph, list(measurements.keys()))



"""This module implements functions for translating to and from Cirq."""
from __future__ import annotations

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import cirq

from bqskit.ir.circuit import Circuit
from bqskit.ir.lang.qasm2 import OPENQASM2Language


def cirq_to_bqskit(cc: cirq.Circuit) -> Circuit:
    """Convert Cirq's Circuit `cc` to a BQSKit Circuit."""
    try:
        import cirq
    except ImportError as e:
        raise ImportError(
            '\n\nUnable to import cirq package.\n'
            'Please ensure that it is installed with the following command:\n'
            '\tpip install cirq\n',
        ) from e

    circuit = OPENQASM2Language().decode(cirq.qasm(cc))
    return circuit


def bqskit_to_cirq(circuit: Circuit) -> cirq.Circuit:
    """Convert a BQSKit Circuit to Cirq's Circuit."""
    try:
        from cirq.contrib.qasm_import import circuit_from_qasm
    except ImportError as e:
        raise ImportError(
            '\n\nUnable to import cirq package.\n'
            'Please ensure that it is installed with the following command:\n'
            '\tpip install cirq\n',
        ) from e

    return circuit_from_qasm(OPENQASM2Language().encode(circuit))



"""This module implements the TrivialPlacementPass class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
_logger = logging.getLogger(__name__)


class TrivialPlacementPass(BasePass):
    """Place the logical qubits on the first n physical qubits."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Select the first n physical qubits for placement
        trivial_placement = list(range(circuit.num_qudits))
        model = BasePass.get_model(circuit, data)
        data['placement'] = trivial_placement

        _logger.info(f'Placed qudits on {data["placement"]}')

        # Raise an error if this is not a valid placement
        sg = model.coupling_graph.get_subgraph(data['placement'])
        if not sg.is_fully_connected():
            raise RuntimeError('The trivial placement is not valid.')



"""This module implements the CompressPass class."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit


class CompressPass(BasePass):
    """
    The CompressPass class.

    The CompressPass compresses the circuit's cycles which might make future
    passes slightly more efficient without actually changing the circuit.
    """

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        circuit.compress()



"""This module implements the PAMRoutingPass class."""
from __future__ import annotations

import itertools as it
import logging
from typing import Dict
from typing import Literal
from typing import overload
from typing import Sequence
from typing import Tuple
from typing import TypedDict

import numpy as np

from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.gates.constant.swap import SwapGate
from bqskit.ir.point import CircuitPoint
from bqskit.passes.mapping.sabre import GeneralizedSabreAlgorithm
from bqskit.qis.graph import CouplingGraph
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

_logger = logging.getLogger(__name__)


PAMBlockPermData = Dict[Tuple[Tuple[int, ...], Tuple[int, ...]], Circuit]
PAMBlockTAPermData = Dict[CouplingGraph, PAMBlockPermData]


class PAMBlockResultData(TypedDict):
    pre_perm: tuple[int, ...]
    post_perm: tuple[int, ...]
    original_utry: UnitaryMatrix


PAMBlockResultDict = Dict[CircuitPoint, PAMBlockResultData]


class PermutationAwareMappingAlgorithm(GeneralizedSabreAlgorithm):
    """Route the circuit with permutation awareness."""

    def __init__(
        self,
        gate_count_weight: float = 0.1,
        decay_delta: float = 0.001,
        decay_reset_interval: int = 5,
        decay_reset_on_gate: bool = True,
        extended_set_size: int = 20,
        extended_set_weight: float = 0.5,
    ) -> None:
        """
        Construct a PermutationAwareMappingAlgorithm.

        Args:
            gate_count_weight (float): The weight on block gate count
                versus mapping score when selecting a permutation.

            decay_delta (float): The amount to adjust the decay factor by
                each time a swap is applied. Set to zero to disable decay.
                (Default: 0.001)

            decay_reset_interval (int): The amount of swaps to apply before
                reseting the decay factors. (Default: 5)

            decay_reset_on_gate (bool): If true, reset decay factors when
                a logical gate is applied. (Default: True)

            extended_set_size (int): The size of the look-ahead or extended
                set. Set to zero to disable look ahead. (Default: 20)

            extended_set_weight (float): The weight on the extended set
                term when scoring potential swaps. (Default: 0.5)
        """
        if not isinstance(gate_count_weight, float):
            bad_type = type(gate_count_weight)
            m = f'Expected float for gate_count_weight, got {bad_type}'
            raise TypeError(m)

        self.gate_count_weight = gate_count_weight

        super().__init__(
            decay_delta,
            decay_reset_interval,
            decay_reset_on_gate,
            extended_set_size,
            extended_set_weight,
        )

    @overload  # type: ignore
    def forward_pass(
        self,
        circuit: Circuit,
        pi: list[int],
        cg: CouplingGraph,
        perm_data: dict[CircuitPoint, PAMBlockTAPermData],
        modify_circuit: Literal[False] = False,
    ) -> None:
        ...

    @overload
    def forward_pass(
        self,
        circuit: Circuit,
        pi: list[int],
        cg: CouplingGraph,
        perm_data: dict[CircuitPoint, PAMBlockTAPermData],
        modify_circuit: Literal[True],
    ) -> PAMBlockResultDict:
        ...

    def forward_pass(  # type: ignore
        self,
        circuit: Circuit,
        pi: list[int],
        cg: CouplingGraph,
        perm_data: dict[CircuitPoint, PAMBlockTAPermData],
        modify_circuit: bool = False,
    ) -> PAMBlockResultDict | None:
        """
        Apply a forward pass of the PAM algorithm to `pi`.

        Args:
            circuit (Circuit): The circuit to pass over.

            pi (list[int]): The input logical-to-physical mapping. This
                maps logical qudits to physical qudits. So, `pi[l] == p`
                implies logical qudit `l` is sitting on physical qudit `p`.

            cg (CouplingGraph): The connectivity of the hardware.

            perm_data (dict[CircuitPoint, PAMBlockTAPermData]):
                Maps each permutation configuration for every block.

            modfiy_circuit (bool): Whether to modify the circuit as the
                pass is applied or not. (Default: False)
        """
        # Preprocessing
        D = cg.all_pairs_shortest_path()
        F = circuit.front
        decay = [1.0 for i in range(circuit.num_qudits)]
        iter_count = 0
        prev_executed_counts: dict[CircuitPoint, int] = {n: 0 for n in F}
        leading_swaps: list[tuple[int, int]] = []
        _logger.debug(f'Starting forward pam pass with pi: {pi}.')

        if modify_circuit:
            mapped_circuit = Circuit(circuit.num_qudits, circuit.radixes)
            out_data: PAMBlockResultDict = {}

        # Main Loop
        while len(F) > 0:

            # Retrieve executable gates giving the current mapping `pi`
            execute_list = [n for n in F if self._can_exe(circuit[n], pi, cg)]

            # Execute the gates and update F
            if len(execute_list) > 0:
                leading_swaps = []

                for n in execute_list:
                    F.remove(n)
                    prev_executed_counts.pop(n)
                    _logger.debug(f'Executing gate at point {n}.')

                    for successor in circuit.next(n):
                        if successor not in prev_executed_counts:
                            prev_executed_counts[successor] = 1
                        else:
                            prev_executed_counts[successor] += 1
                        num_prev_executed = prev_executed_counts[successor]
                        total_num_prev = len(circuit.prev(successor))
                        if num_prev_executed == total_num_prev:
                            F.add(successor)

                # Permute the qubits on the just executed gates
                E = self._calc_extended_set(circuit, F)
                for n in execute_list:
                    op = circuit[n]

                    if isinstance(op.gate, BarrierPlaceholder):
                        if modify_circuit:
                            physical_location = [pi[q] for q in op.location]
                            mapped_circuit.append_gate(op.gate, op.location)
                        continue

                    p1, circ, p2 = self._get_best_perm(
                        circuit,
                        perm_data[n],
                        cg,
                        F,
                        pi,
                        D,
                        E,
                        op.location,
                    )

                    self._apply_perm(p1, pi)

                    if modify_circuit:
                        physical_location = [pi[q] for q in op.location]
                        cycle = mapped_circuit.append_circuit(
                            circ,
                            physical_location,
                            True,
                        )
                        new_point = CircuitPoint(cycle, physical_location[0])
                        out_data[new_point] = {
                            'pre_perm': self._global_to_local_perm(p1),
                            'post_perm': self._global_to_local_perm(p2),
                            'original_utry': op.get_unitary(),
                        }

                    self._apply_perm(p2, pi)

                # Reset decay if necessary
                if self.decay_reset_on_gate:
                    iter_count = 0
                    for i in range(circuit.num_qudits):
                        decay[i] = 1.0

                continue  # Restart main loop if we executed at least one gate

            # If execute list is empty, check for local-minima
            elif len(leading_swaps) > 5 * cg.num_qudits:
                _logger.debug('Sabre stuck in local minima, backtracking...')

                # Backtrack by removing leading swaps
                for swap in reversed(leading_swaps):
                    self._apply_swap(swap, pi, decay)
                    if modify_circuit:
                        point = mapped_circuit._rear[swap[0]]
                        mapped_circuit.pop(point)
                leading_swaps = []

                # Override heuristic search to progress
                _logger.debug('Overriding sabre search...')
                all_logical_qudits = [circuit[n].location for n in F]
                qudits = min(
                    all_logical_qudits,
                    key=lambda qs: self._get_distance(qs, pi, D),
                )
                for swap in self._uphill_swaps(qudits, cg, pi, D):
                    self._apply_swap(swap, pi, decay)
                    if modify_circuit:
                        mapped_circuit.append_gate(SwapGate(), swap)
                _logger.debug('Stopping override.')
                continue

            # Pick and apply a swap
            E = self._calc_extended_set(circuit, F)
            best_swap = self._get_best_swap(circuit, F, E, D, cg, pi, decay)
            self._apply_swap(best_swap, pi, decay)
            leading_swaps.append(best_swap)

            if modify_circuit:
                mapped_circuit.append_gate(SwapGate(), best_swap)

            # Update loop counter and reset decay if necessary
            iter_count += 1
            if iter_count % self.decay_reset_interval == 0:
                for i in range(circuit.num_qudits):
                    decay[i] = 1.0

        if modify_circuit:
            circuit.become(mapped_circuit)
            return out_data

    def _global_to_local_perm(self, gperm: Sequence[int]) -> tuple[int, ...]:
        """Return the local permutation from a global permutation."""
        global_to_local_map = {q: i for i, q in enumerate(sorted(gperm))}
        return tuple(global_to_local_map[i] for i in gperm)

    def _get_best_perm(
        self,
        circuit: Circuit,
        perm_data: PAMBlockTAPermData,
        cg: CouplingGraph,
        F: set[CircuitPoint],
        pi: list[int],
        D: list[list[float]],
        E: set[CircuitPoint],
        qudits: Sequence[int],
    ) -> tuple[tuple[int, ...], Circuit, tuple[int, ...]]:
        """Return the best permutation to apply before and after a gate."""

        # Local permutations determine how a gate is permuted in it own space
        local_perms = list(it.permutations(range(len(qudits))))

        # Global perms capture local perms' effect on the full logical space
        global_perms = [
            tuple(qudits[i] for i in lperm)
            for lperm in local_perms
        ]

        # Inverted Permutations
        inv_local_perms = [
            tuple(lperm.index(i) for i in range(len(qudits)))
            for lperm in local_perms
        ]

        # Inverted Global Permutations
        inv_global_perms = [
            tuple(qudits[i] for i in ilperm)
            for ilperm in inv_local_perms
        ]

        # Gather valid pre, circ, post triples
        pre_circ_post_triples = []
        perm_iter = zip(local_perms, inv_local_perms, inv_global_perms)
        for lperm, ilperm, gperm1 in perm_iter:
            physical_location = [pi[qudits[p]] for p in ilperm]
            local_graph = cg.get_subgraph(physical_location)
            if local_graph in perm_data:
                for perms, circ in perm_data[local_graph].items():
                    if lperm == perms[0]:
                        gperm2 = global_perms[local_perms.index(perms[1])]
                        pre_circ_post_triples.append((gperm1, circ, gperm2))

        if len(pre_circ_post_triples) == 0:
            raise RuntimeError(
                'Unable to find any valid permutated circuits.\n'
                'You must embed proper permutation aware synthesis results'
                ' first before running this pass.\n'
                'If you are already running an'
                ' EmbedAllPermutationsPass, try toggling topology selection.',
            )

        # For each permutation get the entangling gate count
        mq_gate_counts = []
        sq_gate_counts = []
        for _, circ, _ in pre_circ_post_triples:
            num_tq_gates = 0
            num_sq_gates = 0
            for gate, count in circ.gate_counts.items():
                if gate.num_qudits >= 2:
                    num_tq_gates += count
                else:
                    num_sq_gates += count
            mq_gate_counts.append(num_tq_gates)
            sq_gate_counts.append(num_sq_gates)

        # If no more gates after this one, then pick the shortest circuit
        if len(F) == 0:
            best_index = np.argmin(mq_gate_counts)
            return pre_circ_post_triples[best_index]

        # Calculate best scoring permutation
        best_triple = pre_circ_post_triples[0]
        best_perm = (best_triple[0], best_triple[2])
        mapping_score = self._score_perm(circuit, F, pi, D, best_perm, E)
        gate_score = mq_gate_counts[0] * self.gate_count_weight / len(F)
        best_score = mapping_score + gate_score

        for i in range(1, len(pre_circ_post_triples)):
            gperm = (pre_circ_post_triples[i][0], pre_circ_post_triples[i][2])
            score = self._score_perm(circuit, F, pi, D, gperm, E)
            score = mq_gate_counts[i] * self.gate_count_weight / len(F) + score
            if score < best_score:
                best_score = score
                best_perm = gperm
                best_triple = pre_circ_post_triples[i]
        return best_triple

    def _score_perm(
        self,
        circuit: Circuit,
        F: set[CircuitPoint],
        pi: list[int],
        D: list[list[float]],
        perm: tuple[Sequence[int], Sequence[int]],
        E: set[CircuitPoint],
    ) -> float:
        """Calculating the routing score after applying `perm`."""
        pi_bkp = pi.copy()
        pi_c = {q: pi[perm[0][i]] for i, q in enumerate(sorted(perm[0]))}
        for q in perm[0]:
            pi[q] = pi_c[q]
        pi_c = {q: pi[perm[1][i]] for i, q in enumerate(sorted(perm[1]))}
        for q in perm[1]:
            pi[q] = pi_c[q]

        # Front Set Term
        front = 0.0
        for n in F:
            min_term = np.inf
            for q in circuit[n].location:
                term = 0.0
                for p in circuit[n].location:
                    if p == q:
                        continue
                    term += D[pi[q]][pi[p]]
                min_term = min(term, min_term)
            front += min_term
        front /= len(F)

        # Extended Set Term
        extend = 0.0
        if len(E) > 0:
            for n in E:
                min_term = np.inf
                for q in circuit[n].location:
                    term = 0.0
                    for p in circuit[n].location:
                        if p == q:
                            continue
                        term += D[pi[q]][pi[p]]
                    min_term = min(term, min_term)
                extend += min_term
            extend /= len(E)
            extend *= self.extended_set_weight

        pi[:] = pi_bkp[:]
        return front + extend



"""This module implements the SimpleLayerGenerator class."""
from __future__ import annotations

import logging

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
_logger = logging.getLogger(__name__)


class SimpleLayerGenerator(LayerGenerator):
    """
    The SimpleLayerGenerator class.

    Starts a circuit by placing a single-qudit gate on each qudit. Expands a
    circuit by placing a two-qudit building block on all valid links. Each
    building block is composed of a two-qudit gate followed by two single-qudit
    gates.
    """

    def __init__(
        self,
        two_qudit_gate: Gate = CNOTGate(),
        single_qudit_gate_1: Gate = U3Gate(),
        single_qudit_gate_2: Gate | None = None,
        initial_layer_gate: Gate | None = None,
    ) -> None:
        """
        Construct a SimpleLayerGenerator.

        Args:
            two_qudit_gate (Gate): A two-qudit gate that starts this
                layer generator's building block. (Default: CNOTGate())

            single_qudit_gate_1 (Gate): A single-qudit gate that follows
                `two_qudit_gate` in the building block. (Default: U3Gate())

            single_qudit_gate_2 (Gate | None): An alternate single-qudit
                gate to be used as the second single-qudit gate in the
                building block. If left as None, defaults to
                `single_qudit_gate_1`. (Default: None)

            initial_layer_gate (Gate | None): An alternate single-qudit
                gate that creates the initial layer. If left as None,
                defaults to `single_qudit_gate_1`. (Default: None)

        Raises:
            ValueError: If `two_qudit_gate`'s size is not 2, or if any
                of the single-qudit gates' size is not 1.

            ValueError: If `single_qudit_gate_1`'s radix does not match
                the radix of `two_qudit_gate`'s first qudit, or if
                `single_qudit_gate_2`'s radix does not match the radix
                of `two_qudit_gate`'s second qudit.
        """

        if not isinstance(two_qudit_gate, Gate):
            raise TypeError(
                'Expected gate for two_qudit_gate, got %s.'
                % type(two_qudit_gate),
            )

        if two_qudit_gate.num_qudits != 2:
            raise ValueError(
                'Expected two-qudit gate'
                ', got a gate that acts on %d qudits.'
                % two_qudit_gate.num_qudits,
            )

        if not isinstance(single_qudit_gate_1, Gate):
            raise TypeError(
                'Expected gate for single_qudit_gate_1, got %s.'
                % type(single_qudit_gate_1),
            )

        if single_qudit_gate_1.num_qudits != 1:
            raise ValueError(
                'Expected single-qudit gate'
                ', got a gate that acts on %d qudits.'
                % single_qudit_gate_1.num_qudits,
            )

        if single_qudit_gate_2 is None:
            single_qudit_gate_2 = single_qudit_gate_1

        if initial_layer_gate is None:
            initial_layer_gate = single_qudit_gate_1

        if not isinstance(single_qudit_gate_2, Gate):
            raise TypeError(
                'Expected gate for single_qudit_gate_2, got %s.'
                % type(single_qudit_gate_2),
            )

        if single_qudit_gate_2.num_qudits != 1:
            raise ValueError(
                'Expected single-qudit gate'
                ', got a gate that acts on %d qudits.'
                % single_qudit_gate_2.num_qudits,
            )

        if not isinstance(initial_layer_gate, Gate):
            raise TypeError(
                'Expected gate for initial_layer_gate, got %s.'
                % type(initial_layer_gate),
            )

        if initial_layer_gate.num_qudits != 1:
            raise ValueError(
                'Expected single-qudit gate'
                ', got a gate that acts on %d qudits.'
                % initial_layer_gate.num_qudits,
            )

        two_radix_1 = two_qudit_gate.radixes[0]
        two_radix_2 = two_qudit_gate.radixes[1]

        if two_radix_1 != single_qudit_gate_1.radixes[0]:
            raise ValueError(
                'Radix mismatch between two_qudit_gate and single_qudit_gate_1'
                ': %d != %d.'
                % (two_radix_1, single_qudit_gate_1.radixes[0]),
            )

        if two_radix_2 != single_qudit_gate_2.radixes[0]:
            raise ValueError(
                'Radix mismatch between two_qudit_gate and single_qudit_gate_2'
                ': %d != %d.'
                % (two_radix_2, single_qudit_gate_2.radixes[0]),
            )

        self.two_qudit_gate = two_qudit_gate
        self.single_qudit_gate_1 = single_qudit_gate_1
        self.single_qudit_gate_2 = single_qudit_gate_2
        self.initial_layer_gate = initial_layer_gate

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a radix mismatch with
                `self.initial_layer_gate`.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        for radix in target.radixes:
            if radix != self.initial_layer_gate.radixes[0]:
                raise ValueError(
                    'Radix mismatch between target and initial_layer_gate.',
                )

        init_circuit = Circuit(target.num_qudits, target.radixes)
        for i in range(init_circuit.num_qudits):
            init_circuit.append_gate(self.initial_layer_gate, [i])
        return init_circuit

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits < 2:
            raise ValueError('Cannot expand a single-qudit circuit.')

        # Get the coupling graph
        coupling_graph = data.connectivity

        # Generate successors
        successors = []
        for edge in coupling_graph:
            successor = circuit.copy()
            successor.append_gate(self.two_qudit_gate, [edge[0], edge[1]])
            successor.append_gate(self.single_qudit_gate_1, edge[0])
            successor.append_gate(self.single_qudit_gate_2, edge[1])
            successors.append(successor)

        return successors



"""This module implements a general Diagonal Gate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class DiagonalGate(
    QubitGate,
    CachedClass,
    LocallyOptimizableUnitary,
):
    """
    A gate representing a general diagonal unitary. The top-left element is
    fixed to 1, and the rest are set to exp(i * theta).

    This gate is used to optimize the Block ZXZ decomposition of a unitary.
    """
    _qasm_name = 'diag'

    def __init__(
        self,
        num_qudits: int = 2,
    ):
        self._num_qudits = num_qudits
        # 1 parameter per diagonal element, removing one for global phase
        self._num_params = 2 ** num_qudits - 1

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        mat = np.eye(2 ** self.num_qudits, dtype=np.complex128)

        for i in range(1, 2 ** self.num_qudits):
            mat[i][i] = np.exp(1j * params[i - 1])

        return UnitaryMatrix(mat)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        grad = np.zeros(
            (
                len(params), 2 ** self.num_qudits,
                2 ** self.num_qudits,
            ), dtype=np.complex128,
        )

        for i, ind in enumerate(range(1, 2 ** self.num_qudits)):
            grad[i][ind][ind] = 1j * np.exp(1j * params[i])

        return grad

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        self.check_env_matrix(env_matrix)
        thetas = [0.0] * self.num_params

        base = env_matrix[0, 0]
        if base == 0:
            base = np.max(env_matrix[0, :])

        for i in range(1, 2 ** self.num_qudits):
            # Optimize each angle independently
            a = np.angle(env_matrix[i, i] / base)
            thetas[i - 1] = -1 * a

        return thetas



"""The BQSKit Quantum Compiler Framework Setup File."""
from __future__ import annotations

import os

from setuptools import find_packages
from setuptools import setup

root_dir_path = os.path.abspath(os.path.dirname(__file__))
pkg_dir_path = os.path.join(root_dir_path, 'bqskit')
readme_path = os.path.join(root_dir_path, 'README.md')
version_path = os.path.join(pkg_dir_path, '_version.py')

# Load Version Number
with open(version_path) as version_file:
    exec(version_file.read())

# Load Readme
with open(readme_path) as readme_file:
    long_description = readme_file.read()

setup(
    name='bqskit',
    version=__version__,  # type: ignore # noqa # Defined in version.py loaded above
    description='Berkeley Quantum Synthesis Toolkit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/BQSKit/bqskit',
    author='LBNL - BQSKit developers',
    author_email='edyounis@lbl.gov',
    license='BSD 3-Clause License',
    license_files=['LICENSE'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development :: Compilers',
        'Typing :: Typed',
    ],
    keywords=[
        'berkeley',
        'quantum',
        'synthesis',
        'toolkit',
        'partitioning',
    ],
    project_urls={
        'Bug Tracker': 'https://github.com/BQSKit/bqskit/issues',
        'Source Code': 'https://github.com/BQSKit/bqskit',
        'Documentation': 'https://bqskit.readthedocs.io/en/latest',
    },
    packages=find_packages(exclude=['examples*', 'test*']),
    install_requires=[
        'bqskitrs>=0.4.1',
        'lark',
        'numpy>=1.22.0',
        'scipy>=1.8.0',
        'typing-extensions>=4.0.0',
        'dill>=0.3.8',
    ],
    python_requires='>=3.8, <4',
    entry_points={
        'console_scripts': [
            'bqskit-server = bqskit.runtime.detached:start_server',
            'bqskit-manager = bqskit.runtime.manager:start_manager',
            'bqskit-worker = bqskit.runtime.worker:start_worker_rank',
        ],
    },
    extras_require={
        'dev': [
            'hypothesis[zoneinfo]',
            'mypy',
            'pre-commit',
            'psutil',
            'pytest',
            'tox',
        ],
    },
)



"""This module implements the ScipyMinimizer class."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import minimize

from bqskit.ir.opt.minimizer import Minimizer

if TYPE_CHECKING:
    from bqskit.ir.opt.cost import CostFunction
    from bqskit.qis.unitary.unitary import RealVector
    import numpy.typing as npt


class ScipyMinimizer(Minimizer):
    """A minimizer that uses scipy's 'minimize' function."""

    def __init__(self, tol: float = 1e-10) -> None:
        """
        Construct a ScipyMinimizer by passing all arguments to scipy.

        Args:
            tol (float): Set the tolerance used during scipy's minimize.
        """
        self.tol = tol

    def minimize(
        self,
        cost: CostFunction,
        x0: RealVector,
    ) -> npt.NDArray[np.float64]:
        """Solve the minimization problem by using scipy."""
        if len(x0) == 0:
            return np.array([])
        return minimize(cost, x0, tol=self.tol).x



"""This module implements MachineModels for Rigetti's QPUs."""
from __future__ import annotations

from bqskit.compiler.machine import MachineModel
from bqskit.ir.gates.constant.cz import CZGate
from bqskit.ir.gates.constant.iswap import ISwapGate
from bqskit.ir.gates.constant.sx import SXGate
from bqskit.ir.gates.constant.x import XGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.qis.graph import CouplingGraph

rigetti_gate_set = {SXGate(), XGate(), RZGate(), CZGate()}

ankaa_gate_set = {SXGate(), XGate(), RZGate(), CZGate(), ISwapGate()}

_aspen_11_coupling_graph = CouplingGraph([
    # Ring 1
    (0, 1), (1, 2), (2, 3), (3, 4),
    (4, 5), (5, 6), (6, 7), (7, 0),

    # Ring 2
    (8, 9), (9, 10), (10, 11), (11, 12),
    (12, 13), (13, 14), (14, 15), (15, 8),

    # Ring 3
    (16, 17), (17, 18), (19, 20),
    (20, 21), (21, 22), (22, 23), (23, 16),

    # Ring 4
    (24, 25), (25, 26), (26, 27), (27, 28),
    (28, 29), (29, 30), (30, 31), (31, 24),

    # Ring 5
    (34, 35), (35, 36),
    (36, 37), (37, 38), (38, 39),

    # Ring 1-2
    (1, 14), (2, 13),

    # Ring 2-3
    (9, 22), (10, 21),

    # Ring 3-4
    (17, 30),

    # Ring 4-5
    (25, 38), (26, 37),
])
"""Retrieved August 31, 2022: https://qcs.rigetti.com/qpus."""

_octo_rings = [
    [
        (i + 0, i + 1), (i + 1, i + 2), (i + 2, i + 3), (i + 3, i + 4),
        (i + 4, i + 5), (i + 5, i + 6), (i + 6, i + 7), (i + 7, i + 0),
    ]
    for i in range(0, 10 * 8, 8)
]

_horizontal_connections = [
    [
        (i + 1, j + 6),
        (i + 2, j + 5),
    ]
    for i, j in [(8 * r, 8 * (r + 1)) for r in [0, 1, 2, 3, 5, 6, 7, 8]]
]

_vertical_connections = [
    [
        (i + 0, j + 3),
        (i + 7, j + 4),
    ]
    for i, j in [(8 * r, 8 * (r + 5)) for r in range(5)]
]
_links = []
for l in _octo_rings:
    _links.extend(l)
for l in _horizontal_connections:
    _links.extend(l)
for l in _vertical_connections:
    _links.extend(l)

_aspen_m2_coupling_graph = CouplingGraph(_links)
"""Retrieved August 31, 2022: https://qcs.rigetti.com/qpus."""

_ankaa_9q_3_coupling_graph = CouplingGraph.grid(3, 3)
"""Retrieved September 11, 2024: https://qcs.rigetti.com/qpus."""

_ankaa_2_coupling_graph = CouplingGraph.grid(7, 12)
"""Retrieved September 11, 2024: https://qcs.rigetti.com/qpus."""

Aspen11Model = MachineModel(40, _aspen_11_coupling_graph, rigetti_gate_set)
"""A BQSKit MachineModel for Rigetti's Aspen-11 quantum processor."""

AspenM2Model = MachineModel(80, _aspen_m2_coupling_graph, rigetti_gate_set)
"""A BQSKit MachineModel for Rigetti's Aspen-M-2 quantum processor."""

ANKAA2Model = MachineModel(84, _ankaa_2_coupling_graph, ankaa_gate_set)

ANKAA9Q3Model = MachineModel(9, _ankaa_9q_3_coupling_graph, ankaa_gate_set)

__all__ = ['Aspen11Model', 'AspenM2Model']



"""This module defines the OPENQASM 2.0 Grammer and Lark Parser for it."""
from __future__ import annotations

from lark import Lark
from lark import Tree

_OPENQASMPARSER = Lark(
    r"""
ID: /[a-zA-Z][A-Za-z0-9_]*/
REAL:/([0-9]+(\.[0-9]*)?|\.[0-9]+)([eE][-+]?[0-9]+)?/
NNINTEGER: /[1-9]+[0-9]*|0/
PI: "pi"
SIN: "sin"
COS: "cos"
TAN: "tan"
EXP: "EXP"
LN: "ln"
SQRT: "sqrt"
COMMENT: /\/\/+.*/
mainprogram: "OPENQASM" REAL ";" program
program: statement | program statement
statement: decl
            | gatedecl goplist rbracket
            | gatedecl rbracket
            | "opaque" ID idlist ";"
            | "opaque" ID "(" ")" idlist ";"
            | "opaque" ID "(" idlist ")" idlist ";"
            | qop
            | "if" "(" ID "==" NNINTEGER ")" qop
            | barrier
            | incstmt
incstmt: "include" ESCAPED_STRING ";"
decl: qreg | creg
creg: "creg" ID "[" NNINTEGER "]" ";"
qreg: "qreg" ID "[" NNINTEGER "]" ";"
barrier: "barrier" anylist ";"
barrierp: "barrier" anylist ";"
gatedecl: "gate" ID idlist "{"
            | "gate" ID "(" ")" idlist "{"
            | "gate" ID "(" idlist ")" idlist "{"
goplist: uopp
            | "barrierp" idlist ";"
            | goplist uopp
            | goplist "barrier" idlist ";"
qop: uop
        | measure
        | reset
measure: "measure" argument "->" argument ";"
reset: "reset" argument ";"
uop: ugate
        | cxgate
        | gate
gate: ID anylist ";"
        | ID "(" ")" anylist ";"
        | ID "(" explist ")" anylist ";"
ugate: "U" "(" explist ")" argument ";"
cxgate: "CX" argument "," argument ";"
uopp: ugatep
        | cxgatep
        | gatep
gatep: ID anylist ";"
        | ID "(" ")" anylist ";"
        | ID "(" explist ")" anylist ";"
ugatep: "U" "(" explist ")" argument ";"
cxgatep: "CX" argument "," argument ";"
anylist: idlist | mixedlist
idlist: ID | idlist "," ID
mixedlist: ID "[" NNINTEGER "]" | mixedlist "," ID
                                    | mixedlist "," ID "[" NNINTEGER "]"
                                    | idlist "," ID "[" NNINTEGER "]"
argument: ID | ID "[" NNINTEGER "]"
explist: exp | explist "," exp
exp: mulexp ((/\+/ | /\-/) mulexp)*
mulexp: primaryexp ((/\*/ | /\//) primaryexp)*
usub: "-" exp
pow: primaryexp "^" primaryexp
parenexp: "(" exp ")"
unaryexp: unaryop "(" exp ")"
primaryexp: parenexp | REAL | NNINTEGER | PI | ID | pow | usub | unaryexp
unaryop: SIN
        | COS
        | TAN
        | EXP
        | LN
        | SQRT
rbracket: "}"

%import common.ESCAPED_STRING
%import common.WS
%ignore WS
%ignore COMMENT
""",
    parser='lalr',
    start='mainprogram',
)


def parse(s: str) -> Tree:
    return _OPENQASMPARSER.parse(s)



"""This module implements the WhileLoopPass class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.workflow import Workflow
from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.compiler.passdata import PassData
    from bqskit.compiler.workflow import WorkflowLike

_logger = logging.getLogger(__name__)


class WhileLoopPass(BasePass):
    """
    The WhileLoopPass class.

    This is a control pass that conditionally executes a workflow in a loop.
    """

    def __init__(
        self,
        condition: PassPredicate,
        loop_body: WorkflowLike,
    ) -> None:
        """
        Construct a WhileLoopPass.

        Args:
            condition (PassPredicate): The condition checked.

            loop_body (WorkflowLike): The pass or passes to execute
                while `condition` is true.
        """
        if not isinstance(condition, PassPredicate):
            raise TypeError(f'Expected PassPredicate, got {type(condition)}.')

        self.condition = condition
        self.workflow = Workflow(loop_body)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        while self.condition(circuit, data):
            _logger.debug('Loop body executing...')
            await self.workflow.run(circuit, data)



"""This module implements the StateVectorMap base class."""
from __future__ import annotations

import abc

from bqskit.qis.state.state import StateLike
from bqskit.qis.state.state import StateVector


class StateVectorMap(abc.ABC):
    """A map from quantum pure states to quantum pure states."""

    @abc.abstractmethod
    def get_statevector(self, in_state: StateLike) -> StateVector:
        """
        Calculate the output state given the input state.

        Args:
            in_state (StateLike): The pure quantum state input to the
                state vector map.

        Returns:
            StateVector: The output quantum state.
        """



"""This module implements the ManyQuditGatesPredicate class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class ManyQuditGatesPredicate(PassPredicate):
    """Check if many-qudit gates are in the circuit and/or model."""

    def __init__(
        self,
        check_circuit: bool = True,
        check_model: bool = True,
    ) -> None:
        """
        Construct a ManyQuditGatesPredicate.

        Args:
            check_circuit (bool): Whether to check the circuit for many-qudit
                gates. (Default: True)

            check_model (bool): Whether to check the model for many-qudit
                gates. (Default: True)

        Raises:
            ValueError: If both check_circuit and check_model are False.
        """
        if not isinstance(check_circuit, bool):
            raise TypeError(
                f'Expected bool for check_circuit, got {type(check_circuit)}.',
            )

        if not isinstance(check_model, bool):
            raise TypeError(
                f'Expected bool for check_model, got {type(check_model)}.',
            )

        if not check_circuit and not check_model:
            raise ValueError(
                'At least one of check_circuit or check_model must be True.',
            )

        self.check_circuit = check_circuit
        self.check_model = check_model

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        if self.check_circuit:
            for gate in circuit.gate_set_no_blocks:
                if gate.num_qudits > 2:
                    _logger.debug(f'Found many-qudit gate in circuit: {gate}.')
                    return True

        if self.check_model:
            for gate in data.gate_set:
                if gate.num_qudits > 2:
                    _logger.debug(f'Found many-qudit gate in model: {gate}.')
                    return True

        return False



"""This module implements the ConstantGate base class."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class ConstantGate(
    Gate,
    DifferentiableUnitary,
    LocallyOptimizableUnitary,
    CachedClass,
):
    """A gate that does not change during circuit instantiation."""

    _num_params = 0
    _utry: UnitaryMatrix

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)
        return getattr(self, '_utry')

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)
        return np.array([])

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)
        return getattr(self, '_utry'), np.array([])

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        return []



"""This module implements the MPRYGate."""
from __future__ import annotations

import typing

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


def get_indices(
    index: int,
    target_qudit: int,
    num_qudits: int,
) -> tuple[int, int]:
    """Get indices for the matrix based on the target qubit."""
    shift_qubit = num_qudits - target_qudit - 1
    shift = 2 ** shift_qubit
    # Split into two parts around target qubit
    # 100 | 111
    left = index // shift
    right = index % shift

    # Now, shift left by one spot to
    # make room for the target qubit
    left *= (shift * 2)
    # Now add 0 * new_ind and 1 * new_ind to get indices
    return left + right, left + shift + right


class MPRYGate(
    QubitGate,
    DifferentiableUnitary,
    CachedClass,
    LocallyOptimizableUnitary,
):
    """
    A gate representing a multiplexed Y rotation. A multiplexed Y rotation
    uses n - 1 qubits as select qubits and applies a Y rotation to the target.
    If the target qubit is the last qubit, then the unitary is block diagonal.
    Each block is a 2x2 RY matrix with parameter theta.

    Since there are n - 1 select qubits, there are 2^(n-1) parameters (thetas).

    We allow the target qubit to be specified to any qubit, and the other qubits
    maintain their order. Qubit 0 is the most significant qubit.

    See this paper: https://arxiv.org/pdf/quant-ph/0406176
    """

    _qasm_name = 'mpry'

    def __init__(
        self,
        num_qudits: int,
        target_qubit: int = -1,
    ) -> None:
        self._num_qudits = num_qudits
        # 1 param for each configuration of the selec qubits
        self._num_params = 2 ** (num_qudits - 1)
        # By default, the controlled qubit is the last qubit
        if target_qubit == -1:
            target_qubit = num_qudits - 1
        self.target_qubit = target_qubit
        super().__init__()

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        matrix = np.zeros(
            (
                2 ** self.num_qudits,
                2 ** self.num_qudits,
            ), dtype=np.complex128,
        )
        for i, param in enumerate(typing.cast(typing.Sequence[float], params)):
            cos = np.cos(param / 2)
            sin = np.sin(param / 2)

            # Now, get indices based on target qubit.
            # i corresponds to the configuration of the
            # select qubits (e.g 5 = 101). Now, the
            # target qubit is 0,1 for both the row and col
            # indices. So, if i = 5 and the target_qubit is 2
            # Then the rows/cols are 1001 and 1101
            x1, x2 = get_indices(i, self.target_qubit, self.num_qudits)

            matrix[x1, x1] = cos
            matrix[x2, x2] = cos
            matrix[x2, x1] = sin
            matrix[x1, x2] = -1 * sin

        return UnitaryMatrix(matrix)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        grad = np.zeros(
            (
                len(params), 2 ** self.num_qudits,
                2 ** self.num_qudits,
            ), dtype=np.complex128,
        )

        # For each parameter, calculate the derivative
        # with respect to that parameter
        for i, param in enumerate(typing.cast(typing.Sequence[float], params)):
            dcos = -np.sin(param / 2) / 2
            dsin = -1j * np.cos(param / 2) / 2

            # Again, get indices based on target qubit.
            x1, x2 = get_indices(i, self.target_qubit, self.num_qudits)

            grad[i, x1, x1] = dcos
            grad[i, x2, x2] = dcos
            grad[i, x2, x1] = dsin
            grad[i, x1, x2] = -1 * dsin

        return grad

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        self.check_env_matrix(env_matrix)
        thetas: list[float] = [0] * self.num_params

        for i in range(self.num_params):
            x1, x2 = get_indices(i, self.target_qubit, self.num_qudits)
            a = np.real(env_matrix[x1, x1] + env_matrix[x2, x2])
            b = np.real(env_matrix[x2, x1] - env_matrix[x1, x2])
            theta = 2 * np.arccos(a / np.sqrt(a ** 2 + b ** 2))
            theta *= -1 if b > 0 else 1
            thetas[i] = theta

        return thetas

    @staticmethod
    def get_decomposition(params: RealVector = []) -> tuple[
        RealVector,
        RealVector,
    ]:
        """
        Get the corresponding parameters for one level of decomposition of a
        multiplexed gate.

        This is used in the decomposition of both the MPRY and MPRZ gates.
        """
        new_num_params = len(params) // 2
        left_params = np.zeros(new_num_params)
        right_params = np.zeros(new_num_params)
        for i in range(len(left_params)):
            left_param = (params[i] + params[i + new_num_params]) / 2
            right_param = (params[i] - params[i + new_num_params]) / 2
            left_params[i] = left_param
            right_params[i] = right_param

        return left_params, right_params

    @property
    def name(self) -> str:
        """The name of this gate, with the number of qudits appended."""
        base_name = getattr(self, '_name', self.__class__.__name__)
        return f'{base_name}_{self.num_qudits}'



"""This module implements the Rebase2QuditGatePass."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.passes.retarget.two import Rebase2QuditGatePass
from bqskit.qis.unitary import UnitaryMatrix
from bqskit.runtime import get_runtime
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number
_logger = logging.getLogger(__name__)


class AutoRebase2QuditGatePass(Rebase2QuditGatePass):
    """
    The AutoRebase2QuditGatePass class.

    Will use instantiation to change the a 2-qudit gate to a different one.
    """

    def __init__(
        self,
        max_depth: int = 3,
        max_retries: int = -1,
        success_threshold: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        Construct a AutoRebase2QuditGatePass.

        Args:
            max_depth (int): The maximum number of new gates to replace
                an old gate with. (Default: 3)

            max_retries (int): The number of retries for the same gate
                before we increase the maximum depth. If left as -1,
                then never increase max depth. (Default: -1)

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the hilbert schmidt cost function.
                (Default: 1e-8)

            cost (CostFunction | None): The cost function that determines
                successful removal of a gate.
                (Default: HilbertSchmidtResidualsGenerator())

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

        Raises:
            ValueError: if `max_depth` is nonnegative.
        """
        if not is_integer(max_depth):
            raise TypeError(f'Expected integer, got {max_depth}.')

        if max_depth < 0:
            raise ValueError(f'Expected nonnegative depth, got: {max_depth}.')

        if not is_integer(max_retries):
            raise TypeError(f'Expected integer, got {max_retries}.')

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options, got %s.'
                % type(instantiate_options),
            )
        self.max_depth = max_depth
        self.max_retries = max_retries
        self.success_threshold = success_threshold
        self.cost = cost
        self.instantiate_options: dict[str, Any] = {
            'dist_tol': self.success_threshold,
            'min_iters': 100,
        }
        self.instantiate_options.update(instantiate_options)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        new_gates = [g for g in data.gate_set if g.num_qudits == 2]
        old_gates = [
            g for g in circuit.gate_set_no_blocks
            if g not in new_gates and g.num_qudits == 2
        ]
        circs, counts, overdrive, msgs = self.generate_new_gate_templates(
            new_gates,
            data.gate_set.get_general_sq_gate(),
        )

        instantiate_options = self.instantiate_options.copy()
        if 'seed' not in instantiate_options:
            instantiate_options['seed'] = data.seed
        _logger.debug(f'Rebasing gates from {old_gates} to {new_gates}.')

        target = self.get_target(circuit, data)

        if isinstance(target, UnitaryMatrix):
            identity = UnitaryMatrix.identity(target.dim, target.radixes)
            if target.get_distance_from(identity) < self.success_threshold:
                _logger.debug('Target is identity, returning empty circuit.')
                circuit.clear()
                return

        for g in old_gates:
            # Track retries to check for no progress
            num_retries = 0
            prev_count = circuit.count(g)

            while g in circuit.gate_set:
                # Change the seed every iteration to prevent stalls
                if instantiate_options['seed'] is not None:
                    instantiate_options['seed'] += 1

                # Check if we made progress from last loop
                gates_left = circuit.count(g)
                if prev_count == gates_left:
                    num_retries += 1
                else:
                    prev_count = gates_left
                    num_retries = 0

                # Group together a 2-qubit block composed of gates from old set
                point = self.group_near_gates(
                    circuit,
                    circuit.point(g),
                    old_gates,
                )
                circuits_with_new_gate = []
                for circ in circs:
                    circuit_copy = circuit.copy()
                    circuit_copy.replace_with_circuit(point, circ)
                    circuits_with_new_gate.append(circuit_copy)

                # If we have exceeded the number of retries, up the max depth
                if self.max_retries >= 0 and num_retries > self.max_retries:
                    _logger.debug('Exceeded max retries, increasing depth.')
                    circuit_copy = circuit.copy()
                    circuit_copy.replace_with_circuit(point, overdrive)
                    circuits_with_new_gate.append(circuit_copy)

                instantiated_circuits = await get_runtime().map(
                    Circuit.instantiate,
                    circuits_with_new_gate,
                    target=target,
                    **instantiate_options,
                )

                dists = [self.cost(c, target) for c in instantiated_circuits]

                # Find the successful circuit with the least gates
                best_index = None
                best_count = self.max_depth + 2
                for i, dist in enumerate(dists):
                    if dist < self.success_threshold:
                        if counts[i] < best_count:
                            best_index = i
                            best_count = counts[i]

                if best_index is None:
                    circuit.unfold(point)
                    continue

                _logger.debug(msgs[best_index])
                circuit.become(instantiated_circuits[best_index])
